import type { Authenticator } from "../Authenticator.js";
import { ValidationError } from "../errors.js";
import { RawSchema } from "../schema/raw/Schema.js";
import { getBinariesFromTransaction, serializeQueryParams } from "../utils.js";
import { RQLBinary } from "./binary/RQLBinary.js";
import { RQLQueryCellRef } from "./query/references.js";
import { Transaction } from "./transaction/Transaction.js";
import {
  RawTransactionResult,
  TransactionResult,
} from "./transaction/TransactionResult.js";

interface ApiOptions {
  authenticator?: Authenticator;
}

/**
 * Type representing valid final values to use as parameters in a RQL query.
 *
 * For references to requests/binaries, see {@link RQLQueryCellRef} and {@link RQLQueryBinaryRef}
 *
 * @category Api
 */
export type RQLQueryParamValue = string | number | boolean | null;

/**
 * Type representing RQL parameters to send along a request.
 *
 * Each key must be referenced in the RQL query.
 *
 * @category Api
 */
export type RQLQueryParams = Record<
  string,
  RQLQueryParamValue | RQLQueryCellRef | RQLBinary
>;

/**
 * Type representing the currently logged user's data as returned by the `/current-user` route
 *
 * @category Api
 */
export interface CurrentUser {
  login: string;
  eid: number;
  dcTitle: string;
}

/**
 * Type representing the body of an error sent by CubicWeb's API.
 *
 * @category Api
 */
type ApiError = {
  title: string;
  message: string;
  data?: Record<string, unknown>;
};

/**
 * Type representing the error data sent to clients when retrieving an API error.
 *
 * It simply adds the HTTP status code in a `status` key to the regular CubicWeb api error,
 * so users can handle HTTP status codes in their code.
 *
 * @category Api
 */
export type ApiErrorResponse = {
  status: number;
} & ApiError;

/**
 * Class used to handle requests to the CubicWeb API.
 *
 * You should never have to create this object yourself.
 * Instead, retrieve this object from the methods in the {@link Client} class.
 *
 * @internal
 */
export default class Api {
  private _apiUrl: string;
  private _version = "v1";
  private _rqlApiUrl: string;
  private _schemaApiUrl: string;
  private _loginApiUrl: string;
  private _logoutApiUrl: string;
  private _currentUserApiUrl: string;
  private _binaryApiUrl: string;
  private _authenticator?: Authenticator;

  /**
   * Class used to handle requests to the CubicWeb API.
   *
   * You should never have to create this object yourself.
   * Instead, retrieve this object from the methods in the {@link Client} class.
   *
   * @param apiUrl The API url. It should not include the API version as it will be automatically added.
   * Example: if you can retrieve the schema with https://example.com/api/v1/schema,
   * use https://example.com/api as `apiUrl`.
   */
  constructor(apiUrl: string, options: ApiOptions) {
    this._apiUrl = apiUrl;
    this._rqlApiUrl = this.getApiRoute("rql");
    this._schemaApiUrl = this.getApiRoute("schema");
    this._loginApiUrl = this.getApiRoute("login");
    this._logoutApiUrl = this.getApiRoute("logout");
    this._currentUserApiUrl = this.getApiRoute("current-user");
    this._binaryApiUrl = this.getApiRoute("binary");
    this._authenticator = options.authenticator;
  }

  get apiUrl() {
    return this._apiUrl;
  }

  private getApiRoute(route: string) {
    return `${this._apiUrl}/${this._version}/${route}`;
  }

  /**
   * Retrieves the CubicWeb instance's JSON schema.
   *
   * @returns The CubicWeb instance's JSON schema.
   */
  async schema(): Promise<RawSchema> {
    return this.nonNullFetchApi<RawSchema>(this._schemaApiUrl, {
      method: "GET",
    });
  }

  /**
   * Tries to log in the user on the configured CubicWeb instance with the given credentials.
   *
   * @param login The user login
   * @param password The user password
   * @returns A promise returning true if the login was successful, false otherwise
   */
  async login(login: string, password: string): Promise<boolean> {
    try {
      await this.fetchApi<null>(this._loginApiUrl, {
        method: "POST",
        body: JSON.stringify({
          login,
          password,
        }),
      });
      return true;
    } catch (_error) {
      return false;
    }
  }

  /**
   * Logs out the current user. Will invalidate any authentication cookie active.
   *
   * @returns A promise which will resolve when the user is properly logged out
   */
  async logout(): Promise<void> {
    await this.fetchApi<null>(this._logoutApiUrl, {
      method: "POST",
    });
  }

  /**
   * Get the current user associated to this client
   *
   * @returns A promise resolving to the current user
   */
  async currentUser(): Promise<CurrentUser> {
    return await this.nonNullFetchApi<CurrentUser>(this._currentUserApiUrl, {
      method: "GET",
    });
  }

  /**
   * Get the binary for the given entity attribute
   *
   * @param eid Entity eid to fetch the binary from
   * @param attribute Attribute on the entity in which the binary is stored
   * @returns A promise resolving to the binary as a Blob
   */
  async binary(eid: number, attribute: string): Promise<Blob> {
    const params = new URLSearchParams({
      eid: eid.toString(),
      attribute,
    });
    return await this.nonNullFetchApi<Blob>(this._binaryApiUrl + params, {
      method: "GET",
    });
  }

  /**
   * Executes the given transaction.
   *
   * If the {@link Transaction} object contains binaries, the request is automatically
   * sent using `multipart/form-data` to allow uploading binaries.
   *
   * @param transaction Transaction to execute
   * @returns A promise returning the transaction result on success
   */
  async rql(transaction: Transaction): Promise<TransactionResult> {
    const binaries = getBinariesFromTransaction(transaction);
    try {
      if (binaries.length > 0) {
        return await this.multipartTransaction(transaction, binaries);
      } else {
        return await this.jsonTransaction(transaction);
      }
    } catch (e) {
      this.handleUserErrors(e);
    }
  }

  private getTransactionParams(transaction: Transaction) {
    return transaction.queries.map((rqlQuery) => ({
      query: rqlQuery.rql,
      params: serializeQueryParams(transaction, rqlQuery.params),
    }));
  }

  private async multipartTransaction(
    transaction: Transaction,
    binaries: RQLBinary[]
  ) {
    const jsonParams = this.getTransactionParams(transaction);
    const formData = new FormData();
    formData.append("queries", JSON.stringify(jsonParams));
    binaries.forEach((binary) => {
      formData.append(binary.uuid, binary.blob);
    });

    const resultSets = await this.nonNullFetchApi<RawTransactionResult>(
      this._rqlApiUrl,
      {
        method: "POST",
        body: formData,
      }
    );
    return new TransactionResult(transaction, resultSets);
  }

  private async jsonTransaction(transaction: Transaction) {
    const resultSets = await this.nonNullFetchApi<RawTransactionResult>(
      this._rqlApiUrl,
      {
        method: "POST",
        body: JSON.stringify(this.getTransactionParams(transaction)),
      }
    );
    return new TransactionResult(transaction, resultSets);
  }

  private handleUserErrors(error: unknown): never {
    if (
      typeof error === "object" &&
      error !== null &&
      "title" in error &&
      typeof error.title === "string" &&
      "message" in error &&
      typeof error.message === "string"
    ) {
      if (error.title === "ValidationError") {
        throw new ValidationError(error.message);
      }
      throw error;
    }
    throw error;
  }

  /**
   * Behaves the same as `fetchAPI` but forbids null responses.
   * If this function detects a null response from the API,
   * the promise will be rejected with a 500 error code.
   *
   * @param input The API route to use
   * @param init Additional request parameters
   * @returns A promise with the non-null fetched result
   */
  private async nonNullFetchApi<R = unknown>(url: string, init: RequestInit) {
    const result = await this.fetchApi<R>(url, init);
    if (result == null) {
      const error: ApiErrorResponse = {
        status: 500,
        title: "Unexpected null response",
        message: "Response was null but a value was expected",
      };
      throw error;
    }
    return result;
  }

  /**
   * Sends a request to the given API route with the `Content-Type` header set and waits for a response.
   * The returned promise will be rejected if the API returns HTTP error codes.
   * The rejected promise will contain an `ApiErrorResponse` object.
   *
   * @param input The API route to use
   * @param init Additional request parameters
   * @returns A promise with the fetched result
   */
  private async fetchApi<R = unknown>(
    url: string,
    init: RequestInit
  ): Promise<R | null> {
    const headers = new Headers(init.headers);

    // Add custom header  used as CSRF protection
    // note: it is only necessary for non GET requests but can also be useful
    // while debugging.
    if (!headers.has("X-Client-Name")) {
      headers.append("X-Client-Name", "XMLHttpRequest");
    }

    // Add a "Content-Type" header whenever a body is passed.
    // The header might already have been set (either by the caller of this
    // function or by and encoder)
    if (
      init.body &&
      !headers.has("Content-Type") &&
      !(init.body instanceof FormData)
    ) {
      headers.append("Content-Type", "application/json;charset=UTF-8");
    }
    init.headers = headers;
    // Authenticate request
    if (this._authenticator) {
      const authenticatedRequest = await this._authenticator.authenticate(
        url,
        init
      );
      url = authenticatedRequest.url;
      init = authenticatedRequest.init;
    }

    // Execute request
    const response = await fetch(url, init);

    // Parse response body and handle errors
    const status = response.status;
    if (status >= 200 && status < 300) {
      if (status === 204) {
        return null;
      }
      try {
        return await response.json();
      } catch (e) {
        console.error(e);
        return null;
      }
    } else {
      let json;
      try {
        json = await response.json();
      } catch (e) {
        console.error(e);
        throw { status, title: "Error", message: "" };
      }
      throw { status, ...json };
    }
  }
}
