import { Authenticator } from "../Authenticator.js";
import { Schema } from "../schema/classes/Schema.js";
import Api, { RQLQueryParams, CurrentUser } from "./Api.js";
import { ResultSet } from "./ResultSet.js";
import { RQLQuery } from "./query/RQLQuery.js";
import { Transaction } from "./transaction/Transaction.js";
import { TransactionResult } from "./transaction/TransactionResult.js";

export interface ClientOptions {
  authenticator?: Authenticator;
}

/**
 * Class used to communicate with a CubicWeb instance's API.
 *
 * @category Api
 */
export default class Client {
  private api: Api;
  /**
   * Class used to communicate with a CubicWeb instance's API.
   *
   * @param apiUrl The API url. It should not include the API version as it will be automatically added.
   * Example: if you can retrieve the schema with https://example.com/api/v1/schema,
   * use https://example.com/api as `apiUrl`.
   */
  constructor(apiUrl: string, options?: ClientOptions) {
    this.api = new Api(apiUrl, options ?? {});
  }

  get apiUrl() {
    return this.api.apiUrl;
  }

  /**
   * Executes the given query and waits for the response.
   *
   * If binaries are given, the request is automatically
   * sent using `multipart/form-data` to allow uploading binaries.
   *
   * @param rql The RQL query to send
   * @param params The additional parameters for the request
   * @returns A promise with the request's result set
   */
  async execute(rql: string, params: RQLQueryParams = {}): Promise<ResultSet> {
    const transaction = new Transaction();
    const query = new RQLQuery(rql, params);
    transaction.push(query);
    const result = await this.executeTransaction(transaction);
    return result.getResultSet(query);
  }

  /**
   * Retrieves the CubicWeb instance's JSON schema.
   *
   * @returns The CubicWeb instance's JSON schema.
   */
  async schema() {
    const schema = await this.api.schema();
    return new Schema(schema);
  }

  /**
   * Tries to log in the user on the configured CubicWeb instance with the given credentials.
   *
   * @param login The user login
   * @param password The user password
   * @returns A promise returning true if the login was successful, false otherwise
   */
  async login(login: string, password: string) {
    return await this.api.login(login, password);
  }

  /**
   * Logs out the current user. Will invalidate any authentication cookie active.
   *
   * @returns A promise which will resolve when the user is properly logged out
   */
  async logout() {
    return await this.api.logout();
  }

  /**
   * Get the user associated to this client
   *
   * @returns A promise resolving to the current user
   */
  async currentUser(): Promise<CurrentUser> {
    return await this.api.currentUser();
  }

  /**
   * Executes the given transaction.
   *
   * If the {@link Transaction} object contains binaries, the request is automatically
   * sent using `multipart/form-data` to allow uploading binaries.
   *
   * @param transaction Transaction to execute
   * @returns A promise returning the transaction result on success
   */
  async executeTransaction(
    transaction: Transaction
  ): Promise<TransactionResult> {
    return await this.api.rql(transaction);
  }

  /**
   * Get the binary for the given entity attribute
   *
   * @param eid Entity eid to fetch the binary from
   * @param attribute Attribute on the entity in which the binary is stored
   * @returns A promise resolving to the binary as a Blob
   */
  async getBinary(eid: number, attribute: string) {
    return await this.api.binary(eid, attribute);
  }
}
