import { AuthenticatedRequest, Authenticator } from "../../Authenticator.js";
import { encodeFormData } from "../../encodeFormData.js";

export interface SignedRequestAuthToken {
  /**
   * id of the token
   */
  id: string;
  /**
   * secret value of the token
   */
  token: string;
}

function bytesToHex(hashBuffer: ArrayBuffer) {
  const hashArray = Array.from(new Uint8Array(hashBuffer)); // convert buffer to byte array
  const hashHex = hashArray
    .map((b) => b.toString(16).padStart(2, "0"))
    .join(""); // convert bytes to hex string
  return hashHex;
}

/**
 * Authenticator allowing to authenticate requests made to a cubicweb instance
 * with the [signedrequest](https://forge.extranet.logilab.fr/cubicweb/cubes/signedrequest) cube (v3.x).
 *
 * @param token A {@link SignedRequestAuthToken}. This token must be enabled.
 */
export class SignedRequestAuthenticator implements Authenticator {
  private cachedCryptoKey?: CryptoKey;

  constructor(private token: SignedRequestAuthToken) {}

  private async digestBody(body: Uint8Array): Promise<string> {
    const contentHash = await crypto.subtle.digest("SHA-512", body);
    const hashHex = bytesToHex(contentHash);
    return hashHex;
  }

  private async encodeBody(
    body: RequestInit["body"]
  ): Promise<{ body: Uint8Array | null; headers: Headers }> {
    if (typeof body === "string") {
      return {
        body: new TextEncoder().encode(body),
        headers: new Headers(),
      };
    } else if (body instanceof FormData) {
      return await encodeFormData(body);
    } else if (body == null) {
      return {
        body: null,
        headers: new Headers(),
      };
    } else {
      throw new Error("Unexpected body type.");
    }
  }

  private async getCryptoKey() {
    if (this.cachedCryptoKey) {
      return this.cachedCryptoKey;
    }

    const textEncoder = new TextEncoder();
    this.cachedCryptoKey = await crypto.subtle.importKey(
      "raw",
      textEncoder.encode(this.token.token),
      {
        name: "HMAC",
        hash: "SHA-512",
      },
      false,
      ["sign"]
    );
    return this.cachedCryptoKey;
  }

  /**
   * Compute headers used to sign the request and encode the content as bytes
   */
  private async prepareRequest(
    url: string,
    init: RequestInit
  ): Promise<{ contentToSign: string; headers: Headers }> {
    const headers = new Headers(init?.headers);
    const { body: encodedBody, headers: encodingHeaders } =
      await this.encodeBody(init.body || null);
    init.body = encodedBody;
    encodingHeaders.forEach((value, name) => headers.append(name, value));

    const bodyBytes = encodedBody ?? new Uint8Array();

    const bodyHash = await this.digestBody(bodyBytes);
    headers.append("Content-SHA512", bodyHash);

    const date = new Date();
    headers.append("X-Cubicweb-Date", date.toUTCString());

    return {
      contentToSign:
        (init?.method ?? "GET") +
        url +
        bodyHash +
        (headers.get("Content-Type") ?? "") +
        (headers.get("X-Cubicweb-Date") ?? ""),
      headers,
    };
  }

  private async signContent(contentToSign: string): Promise<ArrayBuffer> {
    const textEncoder = new TextEncoder();
    return await crypto.subtle.sign(
      "HMAC",
      await this.getCryptoKey(),
      textEncoder.encode(contentToSign)
    );
  }

  async authenticate(
    url: string,
    init: RequestInit
  ): Promise<AuthenticatedRequest> {
    const { contentToSign, headers } = await this.prepareRequest(url, init);
    headers.append(
      "Authorization",
      `Cubicweb ${this.token.id}:${bytesToHex(
        await this.signContent(contentToSign)
      )}`
    );
    return { url, init: { ...init, headers } };
  }
}
