import { v4 as uuidv4 } from "uuid";
import { RQLQueryCellRef, RQLQueryRowRef } from "./references.js";
import { RQLQueryParams } from "../Api.js";

/**
 * Class representing a RQL query to be sent to the API.
 *
 * Queries are identified by a uuid and are independent from transactions.
 *
 * If a query has been used in a transaction, RQLQuery objects can easily retrieve
 * a reference to the query for later resolving the values in a {@link TransactionResult}.
 *
 * @category Api
 */
export class RQLQuery {
  public readonly queryUUID: string;

  /**
   * Class representing a RQL query to be sent to the API.
   *
   * Queries are identified by a uuid and are independent from transactions.
   *
   * If a query has been used in a transaction, RQLQuery objects can easily retrieve
   * a reference to the query for later resolving the values in a {@link TransactionResult}.
   *
   * @param rql The rql query
   * @param params The parameters for the rql query
   *
   * @category Api
   */
  constructor(
    public readonly rql: string,
    public readonly params: RQLQueryParams = {}
  ) {
    this.queryUUID = uuidv4();
  }

  /**
   * @returns A reference to a {@link ResultSetRow} in the result of this query.
   */
  rowRef(row: number): RQLQueryRowRef {
    return new RQLQueryRowRef(this.queryUUID, row);
  }

  /**
   * @returns A reference to a {@link ResultSetCell} in the result of this query.
   */
  cellRef(row: number, column: number | string): RQLQueryCellRef {
    return new RQLQueryCellRef(this.queryUUID, row, column);
  }
}
