import { getQueryIndexInTransaction } from "../../utils.js";
import {
  RawResultSet,
  ResultSet,
  ResultSetCell,
  ResultSetRow,
} from "../ResultSet.js";
import { RQLQuery } from "../query/RQLQuery.js";
import { RQLQueryCellRef, RQLQueryRowRef } from "../query/references.js";
import { Transaction } from "./Transaction.js";

/**
 * Interface representing the json returned by the API when executing
 * a Transaction
 *
 * @category Api
 */
export interface RawTransactionResult {
  readonly result_sets: RawResultSet[];
}

/**
 * Class representing the result of a transaction.
 *
 * Retrieving the values is done using query references.
 *
 * @example
 *
 * Retrieving the first value in a result set
 *
 * ```tsx
 * const transaction = new Transaction()
 * const query = new RQLQuery("Any X WHERE X name 'test'", {})
 * transaction.push(query)
 * const result = client.executeTransaction(transaction)
 * const eid = result.resolve(query.cellRef(0, 0))
 * ```
 *
 * @category Api
 */
export class TransactionResult {
  private rawResultSets: RawResultSet[];

  constructor(
    private transaction: Transaction,
    rawTransactionResult: RawTransactionResult
  ) {
    this.rawResultSets = rawTransactionResult.result_sets;
  }

  getResultSet(query: RQLQuery): ResultSet {
    const index = getQueryIndexInTransaction(this.transaction, query.queryUUID);
    const resultSet = new ResultSet(this.rawResultSets[index]);
    return resultSet;
  }

  getRow(ref: RQLQueryRowRef): ResultSetRow {
    const index = getQueryIndexInTransaction(this.transaction, ref.queryUUID);
    const resultSet = new ResultSet(this.rawResultSets[index]);
    return resultSet.getRow(ref.row);
  }

  getCell(ref: RQLQueryCellRef): ResultSetCell {
    const index = getQueryIndexInTransaction(this.transaction, ref.queryUUID);
    const resultSet = new ResultSet(this.rawResultSets[index]);
    return resultSet.getCell(ref.row, ref.column);
  }
}
