// Adapted from https://github.com/axios/axios/blob/c6cce43cd94489f655f4488c5a50ecaf781c94f2/lib/helpers/formDataToStream.js#L92
// See https://andreubotella.github.io/multipart-form-data/ for spec and algorithm description

const ALPHA = "abcdefghijklmnopqrstuvwxyz";
const BOUNDARY_ALPHABET = ALPHA + ALPHA.toUpperCase() + "0123456789" + "-_";

const textEncoder = new TextEncoder();

const CRLF = "\r\n";
const CRLF_BYTES = textEncoder.encode(CRLF);
const CRLF_BYTES_COUNT = 2;

class FormDataPart {
  private headers: Uint8Array;
  private contentLength: number;
  public readonly size: number;
  name: string;
  value: Uint8Array | File;
  constructor(name: string, value: FormDataEntryValue) {
    const { escapeName } = FormDataPart;

    let headers = `Content-Disposition: form-data; name="${escapeName(name)}"${
      !(typeof value === "string") && value.name
        ? `; filename="${escapeName(value.name)}"`
        : ""
    }${CRLF}`;
    let bytesOrFile: Uint8Array | File;
    if (typeof value === "string") {
      bytesOrFile = textEncoder.encode(
        String(value).replace(/\r?\n|\r\n?/g, CRLF)
      );
      this.contentLength = bytesOrFile.byteLength;
    } else {
      bytesOrFile = value;
      headers += `Content-Type: ${
        value.type || "application/octet-stream"
      }${CRLF}`;
      this.contentLength = bytesOrFile.size;
    }

    this.headers = textEncoder.encode(headers + CRLF);

    this.size = this.headers.byteLength + this.contentLength + CRLF_BYTES_COUNT;

    this.name = name;
    this.value = bytesOrFile;
  }

  async *encode(): AsyncGenerator<Uint8Array> {
    yield this.headers;

    const { value } = this;

    if (value instanceof Uint8Array) {
      yield value;
    } else {
      yield new Uint8Array(await value.arrayBuffer());
    }

    yield CRLF_BYTES;
  }

  static escapeName(name: string): string {
    return String(name).replace(
      /[\r\n"]/g,
      (match) =>
        ({
          "\r": "%0D",
          "\n": "%0A",
          '"': "%22",
        }[match as "\r" | "\n" | '"'])
    );
  }
}

function generateBoundary(): string {
  let str = "";
  const { length } = BOUNDARY_ALPHABET;
  let size = 25;
  while (size--) {
    str += BOUNDARY_ALPHABET[(Math.random() * length) | 0];
  }

  return str;
}

export async function encodeFormData(body: FormData) {
  const tag = "form-data-boundary";
  const boundary = tag + "-" + generateBoundary();

  const boundaryBytes = textEncoder.encode("--" + boundary + CRLF);
  const footerBytes = textEncoder.encode("--" + boundary + "--" + CRLF + CRLF);
  let contentLength = footerBytes.byteLength;

  const parts = Array.from(body.entries()).map(([name, value]) => {
    const part = new FormDataPart(name, value);
    contentLength += part.size;
    return part;
  });

  contentLength += boundaryBytes.byteLength * parts.length;

  const computedHeaders = new Headers({
    "Content-Type": `multipart/form-data; boundary=${boundary}`,
  });

  if (Number.isFinite(contentLength)) {
    computedHeaders.append("Content-Length", contentLength.toString());
  }

  const fullContent = new Uint8Array(contentLength);
  let offset = 0;

  for (const part of parts) {
    fullContent.set(boundaryBytes, offset);
    offset += boundaryBytes.length;

    for await (const arr of part.encode()) {
      fullContent.set(arr, offset);
      offset += arr.length;
    }
  }
  fullContent.set(footerBytes, offset);

  return { body: fullContent, headers: computedHeaders };
}
