import { EntityRawSchemaArray } from "../raw/Entities.js";
import {
  Cardinality,
  RelationDefinitionRawSchema,
  RelationDefinitionRawSchemaArray,
} from "../raw/RelationDefinitions.js";
import { Schema } from "./Schema.js";
import { EntitySchema } from "./EntitySchema.js";

// This prefix is often applied to relation types as convention
// indicating a relation is viewed from the object's point of view
// We do not export it and instead encourage users to use the utility functions
const REVERSE_INDICATOR = "reverse_";

/**
 * An entity role in a relation.
 *
 * @category Utils
 */
export type EntityRole = "subject" | "object";

/**
 * Returns the given schema using loose typing.
 * Can be useful when working with generic schemas
 *
 * @param schema The schema to create a loose type for
 * @returns A loosely typed schema
 *
 * @category Utils
 */
export const getLooseSchema = <
  E extends EntityRawSchemaArray = EntityRawSchemaArray,
  R extends RelationDefinitionRawSchemaArray<E> = RelationDefinitionRawSchemaArray<E>
>(
  schema: Schema<E, R>
) => {
  return schema as unknown as Schema;
};

/**
 * Checks if the given cardinality requires at least one entity to be present
 *
 * @param cardinality The cardinality to check
 * @returns True if the cardinality is 1 or +
 *
 * @category Utils
 */
export function isCardinalityRequired(cardinality: Cardinality): boolean {
  return cardinality === "1" || cardinality === "+";
}

/**
 * Checks if the cardinality allows multiple entities to be present
 *
 * @param cardinality The cardinility to check
 * @returns True if the cardinality is `*` or `+`
 *
 * @category Utils
 */
export const isCardinalityMultiple = (cardinality: Cardinality): boolean => {
  return cardinality === "*" || cardinality === "+";
};

/**
 * Checks if the cardinality allows entities to be absent
 *
 * @param cardinality The cardinility to check
 * @returns True if the cardinality is `?` or `*`
 *
 * @category Utils
 */
export const isCardinalityOptional = (cardinality: Cardinality): boolean => {
  return cardinality === "?" || cardinality === "*";
};

/**
 * Reverses the given role name
 *
 * @param role The role to reverse
 * @returns "object" if the given role was "subject", "subject" otherwise
 *
 * @category Utils
 */
export function getOppositeRelationRole(role: EntityRole): EntityRole {
  return role === "subject" ? "object" : "subject";
}

/**
 * Checks if the given relation type is reversed.
 *
 * A reversed relation means the relation is considered from the object entity
 * point of view.
 * The reverse indicator is a prefix added as a convention in CubicWeb.
 *
 * @param relationType The relation type to check
 * @returns True if the relation type starts with the reverse indicator
 *
 * @category Utils
 */
export const isRelationTypeReversed = (relationType: string): boolean => {
  return relationType.startsWith(REVERSE_INDICATOR);
};

/**
 * Gets the entity role in the given relation type by
 * checking for the reverse indicator
 *
 * A reversed relation means the relation is considered from the object entity
 * point of view.
 * The reverse indicator is a prefix added as a convention in CubicWeb.
 *
 * @param relationType The relation type to check
 * @returns "subject" if the relation is not reversed, "object" otherwise
 *
 * @category Utils
 */
export const getRoleFromRelationType = (relationType: string): EntityRole => {
  return isRelationTypeReversed(relationType) ? "object" : "subject";
};

/**
 * Adds the reverse indicator from the relation type if it is not already present.
 *
 * A reversed relation means the relation is considered from the object entity
 * point of view.
 * The reverse indicator is a prefix added as a convention in CubicWeb.
 *
 * @param relationType The relation type to transform
 * @returns The relation type with the reverse indicator added or removed
 *
 * @category Utils
 */
export const getReverseRelationType = (relationType: string): string => {
  if (!isRelationTypeReversed(relationType)) {
    return `${REVERSE_INDICATOR}${relationType}`;
  } else {
    return relationType;
  }
};

/**
 * Clears the reverse indicator from the relation type if it is present.
 *
 * A reversed relation means the relation is considered from the object entity
 * point of view.
 * The reverse indicator is a prefix added as a convention in CubicWeb.
 *
 * @param relationType The relation type to transform
 * @returns The relation type with the reverse indicator added or removed
 *
 * @category Utils
 */
export const getRawRelationType = (relationType: string): string => {
  if (isRelationTypeReversed(relationType)) {
    return relationType.replace(REVERSE_INDICATOR, "");
  } else {
    return relationType;
  }
};

/**
 * Returns the relation type with the reverse indicator added if necessary.
 *
 * A reversed relation means the relation is considered from the object entity
 * point of view.
 * The reverse indicator is a prefix added as a convention in CubicWeb.
 *
 * @param relationType The relation type to transform
 * @param role The entity role in the relation
 * @returns A relation type with the reverse indicator if necessary
 *
 * @category Utils
 */
export const getRelationType = (relationType: string, role: EntityRole) => {
  return role === "subject"
    ? relationType
    : getReverseRelationType(relationType);
};

/**
 * Finds the relation matching the provided source, relation typ and target
 *
 * @param entitySchema The entity schema the relation is from
 * @param sourceType The entity type acting as source for the relation
 * @param relationType The type of the relation to use. Can have the reverse indicator to decide the role of the source/target entities
 * @param targetType The entity type acting as target for the relation
 * @returns The relation definition matching the triplet source relation target
 *
 * @category Utils
 */
export const getRelationDefinition = (
  entitySchema: EntitySchema,
  sourceType: string,
  relationType: string,
  targetType: string
) => {
  const role = getRoleFromRelationType(relationType);
  const match = entitySchema.matchRelationDefinitions({
    [role]: sourceType,
    type: getRawRelationType(relationType),
    [getOppositeRelationRole(role)]: targetType,
  });
  if (match.length > 0) {
    return match[0];
  }
  return undefined;
};

/**
 * Partial raw relation definition schema type.
 *
 * @category Utils
 */
export type RelationDefinitionMatcher<E extends EntityRawSchemaArray> = Partial<
  RelationDefinitionRawSchema<E>
>;

/**
 * Type representing a raw relation definition matched from a partial raw relation definition schema.
 * If it finds a relation definition matching the partial one given, it will return its type.
 * If it cannot find any relation definition matching, it will return the `never` type
 * and TypeScript will give compile errors.
 *
 * @category Utils
 */
export type MatchedRelationDefinition<
  E extends EntityRawSchemaArray,
  R extends RelationDefinitionRawSchemaArray<EntityRawSchemaArray>,
  M extends RelationDefinitionMatcher<E>
> = R[number] extends infer RDef ? (RDef extends M ? RDef : never) : never;

export function hasAttributeWithName(
  entitySchema: EntitySchema,
  attributeNameSearched: string
): boolean {
  for (const entityAttribute of entitySchema.attributes) {
    if (entityAttribute.type == attributeNameSearched) {
      return true;
    }
  }
  return false;
}

export function isAttributeRichString(
  entitySchema: EntitySchema,
  attributeNameSearched: string
): boolean {
  return hasAttributeWithName(entitySchema, attributeNameSearched + "_format");
}
