/**
 * Base type used to describe a constraint.
 */
type ConstraintBase<
  T extends string,
  V extends Record<string, unknown> | null
> = {
  type: T;
  value: V;
};

/**
 * The type of a constraint message when one is present.
 */
type ConstraintMessage = string | null;

/**
 * A Unique constraint on an attribute.
 */
export type UniqueConstraint = ConstraintBase<
  "UniqueConstraint",
  { msg: ConstraintMessage }
>;

/**
 * A constraint on the size of a string
 */
export type SizeConstraint = ConstraintBase<
  "SizeConstraint",
  { min: number | null; max: number | null; msg: ConstraintMessage }
>;

/**
 * A Regexp constraint on a string
 */
export type RegexpConstraint = ConstraintBase<
  "RegexpConstraint",
  { regexp: string; flags: number; msg: ConstraintMessage }
>;

/**
 * Reference to an attribute used in constraints.
 *
 * @property \_\_attribute\_\_ name of the attribute in the same entity
 *
 * @category Schema
 */
export interface Attribute {
  __attribute__: string;
}

export interface TimeDelta {
  days: number;
  seconds: number;
  microseconds: number;
}

/**
 * Represents a date relative to the current date at the time of the check.
 *
 * @property \_\_today\_\_ always true
 * @property offset delta to current date at the time of the check
 * @property type name of the Python class used to do the comparison
 *
 * @category Schema
 */
export interface Today {
  __today__: true;
  offset: TimeDelta | null;
  type: "Datetime" | "TZDatetime" | "Date";
}

/**
 * Represents a datetime relative to the current datetime at the time of the check.
 *
 * @property \_\_now\_\_ always true
 * @property offset delta to current date at the time of the check
 *
 * @category Schema
 */
export interface Now {
  __now__: true;
  offset: TimeDelta | null;
}

/**
 * All the possible operators inside a constraint.
 *
 * @category Schema
 */
export type ConstraintOperator = ">" | ">=" | "<" | "<=";

/**
 * A boundary constraint on a number attribute.
 */
export type BoundaryConstraint = ConstraintBase<
  "BoundaryConstraint",
  {
    boundary: Attribute | Today | Now | number | string | null;
    op: ConstraintOperator;
    msg: ConstraintMessage;
  }
>;

/**
 * An interval constraint on a number attribute.
 */
export type IntervalBoundConstraint = ConstraintBase<
  "IntervalBoundConstraint",
  {
    minvalue: number | string | null;
    maxvalue: number | string | null;
    msg: ConstraintMessage;
  }
>;

/**
 * A static vocabulary constraint on a string attribute.
 */
export type StaticVocabularyConstraint = ConstraintBase<
  "StaticVocabularyConstraint",
  { values: string[]; msg: ConstraintMessage }
>;

/**
 * A format constraint on a string attribute.
 */
export type FormatConstraint = ConstraintBase<
  "FormatConstraint",
  {
    values: ["text/rest", "text/markdown", "text/html", "text/plain"];
    msg: ConstraintMessage;
  }
>;

/**
 * A vocabulary constraint on a string attribute described by a RQL.
 */
export type RQLVocabularyConstraint = ConstraintBase<
  "RQLVocabularyConstraint",
  { expression: string; mainvars: string[] }
>;

/**
 * A constraint on an attribute described by a RQL.
 */
export type RQLConstraint = ConstraintBase<
  "RQLConstraint",
  { expression: string; mainvars: string[]; msg: ConstraintMessage }
>;

/**
 * Type representing all possible constraints for attributes.
 *
 * @category Schema
 */
export type AttributeConstraint = Readonly<
  | UniqueConstraint
  | SizeConstraint
  | RegexpConstraint
  | BoundaryConstraint
  | IntervalBoundConstraint
  | StaticVocabularyConstraint
  | FormatConstraint
  | RQLConstraint
>;

/**
 * Type representing all possible constraints for relations.
 *
 * @category Schema
 */
export type RelationConstraint = Readonly<
  RQLVocabularyConstraint | RQLConstraint
>;
