import React, { PropsWithChildren } from "react";
import { render, screen } from "@testing-library/react";
import "@testing-library/jest-dom";
import { AttributeConstraint, Schema } from "@cubicweb/client";
import {
  EntityContextProvider,
  useEntityAttribute,
  useEntityAttributeConstraint,
  useEntityEid,
  useEntitySchema,
  useEntityValue,
  useEntityValues,
} from "../../context/EntityContext.js";
import { CubicWebFormData } from "../../hooks/useCubicWebForm.js";

const schema = new Schema({
  entities: [
    {
      type: "Person",
      description: "A",
      final: false,
    },
    {
      type: "String",
      description: "",
      final: true,
    },
  ],
  relations_definitions: [
    {
      type: "name",
      description: "",
      default: null,
      final: true,
      cardinality: "11",
      subject: "Person",
      object: "String",
      constraints: [
        {
          type: "SizeConstraint",
          value: {
            max: 10,
            min: null,
            msg: "",
          },
        },
        {
          type: "StaticVocabularyConstraint",
          value: {
            values: ["test1", "test2"],
            msg: "",
          },
        },
      ],
      options: {},
    },
    {
      type: "description",
      description: "",
      default: null,
      final: true,
      cardinality: "11",
      subject: "Person",
      object: "String",
      constraints: [],
      options: {},
    },
  ],
});

function TestContext({
  children,
  values,
}: PropsWithChildren<{ values?: CubicWebFormData }>) {
  const entitySchema = schema.getEntity("Person");
  if (entitySchema === null) {
    return <></>;
  }
  return (
    <EntityContextProvider schema={entitySchema} values={values}>
      {children}
    </EntityContextProvider>
  );
}

describe("useEntitySchema", () => {
  function TestComponent() {
    const entitySchema = useEntitySchema();
    const displayValue = entitySchema.type;
    return <p data-testid="test-value">{displayValue}</p>;
  }

  test("should find the schema", () => {
    render(
      <TestContext>
        <TestComponent />
      </TestContext>
    );

    expect(screen.getByTestId("test-value")).toHaveTextContent("Person");
  });

  test("should throw an error without schema", () => {
    expect(() =>
      render(
        <EntityContextProvider>
          <TestComponent />
        </EntityContextProvider>
      )
    ).toThrow();
  });
});

describe("useEntityValues", () => {
  function TestComponent() {
    const values = useEntityValues();
    return (
      <p data-testid="test-value">
        {values ? JSON.stringify(values) : "no data"}
      </p>
    );
  }

  it("should get the values", () => {
    const values = {
      name: "hello",
    };
    render(
      <TestContext values={values}>
        <TestComponent />
      </TestContext>
    );

    expect(screen.getByTestId("test-value")).toHaveTextContent(
      JSON.stringify(values)
    );
  });

  it("should get no data", () => {
    render(
      <TestContext>
        <TestComponent />
      </TestContext>
    );

    expect(screen.getByTestId("test-value")).toHaveTextContent("no data");
  });
});

describe("useEntityValue", () => {
  function TestComponent() {
    const value = useEntityValue("name");
    return <p data-testid="test-value">{value ?? "no data"}</p>;
  }

  it("should find the value for an existing key", () => {
    const values = {
      name: "hello",
    };
    render(
      <TestContext values={values}>
        <TestComponent />
      </TestContext>
    );

    expect(screen.getByTestId("test-value")).toHaveTextContent("hello");
  });

  it("should not find the value for a missing key", () => {
    const values = {
      test: "hello",
    };
    render(
      <TestContext values={values}>
        <TestComponent />
      </TestContext>
    );

    expect(screen.getByTestId("test-value")).toHaveTextContent("no data");
  });

  it("should not find the value if no values at all", () => {
    render(
      <TestContext>
        <TestComponent />
      </TestContext>
    );

    expect(screen.getByTestId("test-value")).toHaveTextContent("no data");
  });
});

describe("useEntityEid", () => {
  function TestComponent() {
    const eid = useEntityEid();
    return <p data-testid="test-value">{eid ?? "no data"}</p>;
  }

  it("should find the eid", () => {
    const values = {
      eid: 1227,
    };
    render(
      <TestContext values={values}>
        <TestComponent />
      </TestContext>
    );

    expect(screen.getByTestId("test-value")).toHaveTextContent("1227");
  });

  it("should find not find the eid", () => {
    render(
      <TestContext>
        <TestComponent />
      </TestContext>
    );

    expect(screen.getByTestId("test-value")).toHaveTextContent("no data");
  });
});

describe("useEntityAttribute", () => {
  function TestComponent({ attribute }: { attribute: string }) {
    const nameAttribute = useEntityAttribute(attribute);
    const displayValue = `${nameAttribute.subject.type} - ${nameAttribute.type} - ${nameAttribute.object.type}`;
    return <p data-testid="test-value">{displayValue}</p>;
  }

  it("should find an existing attribute", () => {
    render(
      <TestContext>
        <TestComponent attribute="name" />
      </TestContext>
    );

    expect(screen.getByTestId("test-value")).toHaveTextContent(
      "Person - name - String"
    );
  });

  it("should throw an error on a missing attribute", () => {
    expect(() =>
      render(
        <TestContext>
          <TestComponent attribute="test" />
        </TestContext>
      )
    ).toThrow();
  });
});

describe("useEntityAttributeConstraint", () => {
  function TestComponent({
    attribute,
    constraint,
  }: {
    attribute: string;
    constraint: AttributeConstraint["type"];
  }) {
    const c = useEntityAttributeConstraint(attribute, constraint);
    const displayValue = c !== null ? JSON.stringify(c.value) : "no-data";
    return <p data-testid="test-value">{displayValue}</p>;
  }

  test("should find the size constraint", () => {
    render(
      <TestContext>
        <TestComponent attribute="name" constraint="SizeConstraint" />
      </TestContext>
    );

    expect(screen.getByTestId("test-value")).toHaveTextContent(
      `{"max":10,"min":null,"msg":""}`
    );
  });

  test("should find the vocabulary constraint", () => {
    render(
      <TestContext>
        <TestComponent
          attribute="name"
          constraint="StaticVocabularyConstraint"
        />
      </TestContext>
    );

    expect(screen.getByTestId("test-value")).toHaveTextContent(
      `{"values":["test1","test2"],"msg":""}`
    );
  });

  test("should handle attributes without constraints", () => {
    render(
      <TestContext>
        <TestComponent attribute="description" constraint="SizeConstraint" />
      </TestContext>
    );

    expect(screen.getByTestId("test-value")).toHaveTextContent("no-data");
  });

  test("should throw on unknown attribute", () => {
    expect(() =>
      render(
        <TestContext>
          <TestComponent attribute="test" constraint="SizeConstraint" />
        </TestContext>
      )
    ).toThrow();
  });
});
