import React from "react";
import { render, screen } from "@testing-library/react";
import "@testing-library/jest-dom";
import { Schema } from "@cubicweb/client";
import {
  RelationDefinitionListContextProvider,
  useRelationDefinitionRole,
  useRelationDefinitionSchemaList,
} from "../../context/RelationDefinitionListContext.js";

const schema = new Schema({
  entities: [
    {
      type: "Person",
      description: "A",
      final: false,
    },
    {
      type: "Task",
      description: "",
      final: true,
    },
  ],
  relations_definitions: [
    {
      type: "has_task",
      description: "",
      final: false,
      cardinality: "11",
      subject: "Person",
      object: "Task",
      constraints: [],
      options: {},
    },
  ],
});

function TestContext({
  children,
  hasData = true,
}: React.PropsWithChildren<{ hasData?: boolean }>) {
  const has_task = schema.matchRelationDefinitions({
    type: "has_task",
  })[0];
  return (
    <RelationDefinitionListContextProvider
      role={hasData ? "subject" : undefined}
      relationDefinitionList={hasData ? [has_task] : undefined}
    >
      {children}
    </RelationDefinitionListContextProvider>
  );
}

describe("useRelationDefinitionSchemaList", () => {
  function TestComponent() {
    let displayValue = "";
    try {
      const relationList = useRelationDefinitionSchemaList();
      displayValue = relationList[0].type;
    } catch (e) {
      displayValue = "ERROR";
    }
    return <p data-testid="test-value">{displayValue}</p>;
  }

  it("should find existing relations", () => {
    render(
      <TestContext>
        <TestComponent />
      </TestContext>
    );

    expect(screen.getByTestId("test-value")).toHaveTextContent("has_task");
  });

  it("should not find missing relations", () => {
    render(
      <TestContext hasData={false}>
        <TestComponent />
      </TestContext>
    );

    expect(screen.getByTestId("test-value")).toHaveTextContent("ERROR");
  });
});

describe("useRelationDefinitionRole", () => {
  function TestComponent() {
    let displayValue = "";
    try {
      displayValue = useRelationDefinitionRole();
    } catch (e) {
      displayValue = "ERROR";
    }
    return <p data-testid="test-value">{displayValue}</p>;
  }

  it("should find existing role", () => {
    render(
      <TestContext>
        <TestComponent />
      </TestContext>
    );

    expect(screen.getByTestId("test-value")).toHaveTextContent("subject");
  });

  it("should not find missing role", () => {
    render(
      <TestContext hasData={false}>
        <TestComponent />
      </TestContext>
    );

    expect(screen.getByTestId("test-value")).toHaveTextContent("ERROR");
  });
});
