import {
  CubicWebErrorData,
  getCubicWebErrorData,
  getErrorMessages,
} from "../../util/errors.js";

describe("getCubicWebErrorData", () => {
  describe("valid value", () => {
    it("handle one line error data", () => {
      const errorData = getCubicWebErrorData(
        "12307 (frequency): Le format de l'intervalle de temps est invalide"
      );
      expect(errorData).toHaveLength(1);
      expect(errorData[0].field).toBe("frequency");
      expect(errorData[0].message).toBe(
        "Le format de l'intervalle de temps est invalide"
      );
    });

    it("multi line error data", () => {
      const errorData = getCubicWebErrorData(
        `12307:
* label: Nom appartient à une contrainte d'unicité transgressée
* scheduled_task_of: Réseau appartient à une contrainte d'unicité transgressée
* : certaines relations transgressent une contrainte d'unicité`
      );
      expect(errorData).toHaveLength(3);
      expect(errorData[0].field).toBe("label");
      expect(errorData[0].message).toBe(
        "Nom appartient à une contrainte d'unicité transgressée"
      );
      expect(errorData[1].field).toBe("scheduled_task_of");
      expect(errorData[1].message).toBe(
        "Réseau appartient à une contrainte d'unicité transgressée"
      );
      expect(errorData[2].field).toBe("");
      expect(errorData[2].message).toBe(
        "certaines relations transgressent une contrainte d'unicité"
      );
    });
  });

  describe("invalid value", () => {
    it("handle one line error data", () => {
      const errorData = getCubicWebErrorData(
        "$* label: Nom appartient à une contrainte d'unicité transgressée"
      );
      expect(errorData).toHaveLength(0);
    });

    it("multi line error data", () => {
      const errorData = getCubicWebErrorData(
        `12306:
$* scheduled_task_of: Réseau appartient à une contrainte d'unicité transgressée
$* : certaines relations transgressent une contrainte d'unicité
$12307 (frequency): Le format de l'intervalle de temps est invalide`
      );
      expect(errorData).toHaveLength(0);
    });
  });
});

describe("getErrorMessages", () => {
  it("detects one local field", () => {
    const cubicWebErrorData: CubicWebErrorData[] = [
      { field: "label", message: "test label" },
    ];

    const errorMessages = getErrorMessages(cubicWebErrorData, ["label"]);
    expect(errorMessages.localErrorList).toHaveLength(1);
    expect(errorMessages.localErrorList[0].field).toBe("label");
    expect(errorMessages.localErrorList[0].message).toBe("test label");
    expect(errorMessages.globalErrorList).toHaveLength(0);
  });

  it("detects one global field", () => {
    const cubicWebErrorData: CubicWebErrorData[] = [
      {
        field: "scheduled_task_of",
        message: "test scheduled_task_of message global",
      },
    ];

    const errorMessages = getErrorMessages(cubicWebErrorData, ["label"]);
    expect(errorMessages.localErrorList).toHaveLength(0);
    expect(errorMessages.globalErrorList).toHaveLength(1);
    expect(errorMessages.globalErrorList[0].field).toBe("scheduled_task_of");
    expect(errorMessages.globalErrorList[0].message).toBe(
      "test scheduled_task_of message global"
    );
  });

  it("detects multiple local fields", () => {
    const cubicWebErrorData: CubicWebErrorData[] = [
      { field: "label", message: "test messege label" },
      { field: "frequency", message: "test message frequency" },
    ];

    const errorMessages = getErrorMessages(cubicWebErrorData, [
      "label",
      "frequency",
    ]);
    expect(errorMessages.localErrorList).toHaveLength(2);
    expect(errorMessages.localErrorList[0].field).toBe("label");
    expect(errorMessages.localErrorList[0].message).toBe("test messege label");
    expect(errorMessages.localErrorList[1].field).toBe("frequency");
    expect(errorMessages.localErrorList[1].message).toBe(
      "test message frequency"
    );
    expect(errorMessages.globalErrorList).toHaveLength(0);
  });

  it("detects multiple global fields", () => {
    const cubicWebErrorData: CubicWebErrorData[] = [
      {
        field: "scheduled_task_of",
        message: "test scheduled_task_of message",
      },
      { field: "", message: "test void field message" },
    ];

    const errorMessages = getErrorMessages(cubicWebErrorData, [
      "label",
      "frequency",
    ]);
    expect(errorMessages.localErrorList).toHaveLength(0);
    expect(errorMessages.globalErrorList).toHaveLength(2);
    expect(errorMessages.globalErrorList[0].field).toBe("scheduled_task_of");
    expect(errorMessages.globalErrorList[0].message).toBe(
      "test scheduled_task_of message"
    );
    expect(errorMessages.globalErrorList[1].field).toBe("");
    expect(errorMessages.globalErrorList[1].message).toBe(
      "test void field message"
    );
  });

  it("detects both local and global fields", () => {
    const cubicWebErrorData: CubicWebErrorData[] = [
      { field: "label", message: "test label message" },
      { field: "frequency", message: "test frequency message" },
      {
        field: "scheduled_task_of",
        message: "test scheduled_task_of message",
      },
      { field: "", message: "test void field message" },
    ];

    const errorMessages = getErrorMessages(cubicWebErrorData, [
      "label",
      "frequency",
    ]);
    expect(errorMessages.localErrorList).toHaveLength(2);
    expect(errorMessages.localErrorList[0].field).toBe("label");
    expect(errorMessages.localErrorList[0].message).toBe("test label message");
    expect(errorMessages.localErrorList[1].field).toBe("frequency");
    expect(errorMessages.localErrorList[1].message).toBe(
      "test frequency message"
    );
    expect(errorMessages.globalErrorList).toHaveLength(2);
    expect(errorMessages.globalErrorList[0].field).toBe("scheduled_task_of");
    expect(errorMessages.globalErrorList[0].message).toBe(
      "test scheduled_task_of message"
    );
    expect(errorMessages.globalErrorList[1].field).toBe("");
    expect(errorMessages.globalErrorList[1].message).toBe(
      "test void field message"
    );
  });
});
