import React, { useMemo, useState } from "react";
import {
  Client,
  EntityRawSchemaArray,
  EntitySchema,
  RelationDefinitionRawSchemaArray,
  Schema,
  Transaction,
  TransactionResult,
  getLooseSchema,
  getReverseRelationType,
} from "@cubicweb/client";

import { CubicWebFormData } from "./useCubicWebForm.js";
import {
  generateSelectEntitiesRQL,
  resultSetToEntities,
} from "@cubicweb/rql-generator";

export interface RelationValues {
  eid: number;
  value: string;
  type: string;
}

/**
 * Parameters to describe how to build a query to the API.
 *
 * @category General Hooks
 */
export interface UseGetCubicWebEntityQueryParams<
  E extends EntityRawSchemaArray,
  EntityType extends E[number]["type"]
> {
  /** The entity type you want to fetch. Allows the lib to know which entity schema to return. */
  type: EntityType;
  /** The eid of the entity you want to fetch. */
  eid: number;
  /** Which attributes/relations you want to ignore while fetching the entity. */
  omittedFields?: string[];
  /** True if you want to include object relations. Defaults to false. */
  includeObjectRelations?: boolean;
}

/**
 * Return values for the useGetCubicWebEntity hook.
 *
 * @category General Hooks
 */
export interface UseGetCubicWebEntityData {
  /** An object describing the values fetched. Undefined when values are not available. */
  values: CubicWebFormData | undefined;
  /** True if values are being loaded, false otherwise */
  loading: boolean;
  /** The entity schema of the entity */
  entityTypeSchema: EntitySchema;
}

/**
 *
 * Fetches the values for the entity of the given eid.
 * Along with the loading state, it also returns the entity schema.
 *
 * Like `useCubicWebForm`, you can decide which relations/attributes to ignore.
 *
 * @example
 * Here is an example usage. Please refer to the documentation of
 * [`@cubicweb/client`](https://cubicweb.pages.logilab.fr/cubicwebjs/client/index.html)
 * on how to create the schema and client parameters.
 *
 * As your instance's schema and client will not change in your app,
 * you will usually want to wrap this hook so you don't have to pass those parameters
 * each time.
 *
 * ```ts
 * import {instanceSchema, client} from "./my-data"
 * import {GetCubicWebEntityQueryParams, useGetCubicWebEntity} from "@cubicweb/react-form-utils"
 *
 * function useGetMyAppEntity(queryParams: GetCubicWebEntityQueryParams) {
 *     return useGetCubicWebEntity(
 *         instanceSchema,
 *         client,
 *         queryParams
 *     )
 * }
 *
 * function MyShowComponent() {
 *     const {values, loading, entityTypeSchema} = useGetMyAppEntity(
 *         {
 *             type: "EntityType",
 *             eid: 1,
 *             omittedFields: ["atttribute_I_want_to_ignore"],
 *             includeObjectRelations: true,
 *         }
 *     )
 *     ...
 * }
 * ```
 *
 * @param schema Your instance's schema from [`@cubicweb/client`](https://cubicweb.pages.logilab.fr/cubicwebjs/client/index.html)
 * @param client Your CubicWeb client from [`@cubicweb/client`](https://cubicweb.pages.logilab.fr/cubicwebjs/client/index.html)
 * @param queryParams Parameters to describe how to build the query.
 *
 * @category General Hooks
 */
export function useGetCubicWebEntity<
  // No need to manually set these types
  // These are automatically inferred from the parameters
  E extends EntityRawSchemaArray,
  R extends RelationDefinitionRawSchemaArray<E>,
  EntityType extends E[number]["type"]
>(
  schema: Schema<E, R>,
  client: Client,
  {
    type,
    eid,
    omittedFields,
    includeObjectRelations,
  }: UseGetCubicWebEntityQueryParams<E, EntityType>
): UseGetCubicWebEntityData {
  const [loading, setLoading] = useState(true);
  const [values, setValues] = React.useState<CubicWebFormData | undefined>(
    undefined
  );
  // It is easier to work with a loosely typed schema for generic generation
  const entityTypeSchema = getLooseSchema(schema).getEntity(type);

  if (entityTypeSchema === null) {
    throw new Error(`Could not find entity ${type}`);
  }

  const availableRelations = useAvailableRelations(
    entityTypeSchema,
    omittedFields,
    includeObjectRelations
  );

  React.useEffect(() => {
    setLoading(true);
    const transaction = new Transaction();
    const queries = generateSelectEntitiesRQL(schema, type, {
      where: { eid },
      relations: availableRelations,
    });
    transaction.push(...queries);

    const resolve = (result: TransactionResult) => {
      return resultSetToEntities(
        schema,
        type,
        availableRelations,
        result.getResultSet(queries[0])
      );
    };

    client
      .executeTransaction(transaction)
      .then(resolve)
      .then((data) => {
        setValues(data);
        setLoading(false);
      });
  }, [eid, type, availableRelations]);

  return {
    values,
    loading,
    entityTypeSchema,
  };
}

export function useAvailableRelations(
  entityTypeSchema: EntitySchema,
  omittedFields?: string[],
  includeObjectRelations = false
) {
  return useMemo(
    () => {
      const subjectRelations = [
        ...entityTypeSchema.attributes,
        ...entityTypeSchema.relationDefinitions.subject,
      ]
        .filter((r) => !omittedFields?.includes(r.type))
        .map((r) => r.type);
      const objectRelations = includeObjectRelations
        ? entityTypeSchema.relationDefinitions.object
            .filter(
              (r) => !omittedFields?.includes(getReverseRelationType(r.type))
            )
            .map((r) => r.type)
        : [];

      return [...subjectRelations, ...objectRelations];
    },
    // entityTypeSchema may be different on each render because of how getEntity is implemented
    []
  );
}
