import { generateCreateEntityRQL } from "../create";
import { mockBlogSchema } from "../utils/test_utils";

describe("generateCreateEntityRQL", () => {
  it("Generates a single query without data", () => {
    const { queries } = generateCreateEntityRQL(mockBlogSchema, "Blog");
    expect(queries).toHaveLength(1);
    const query = queries[0];
    expect(query.rql).toBe("INSERT Blog X");
    expect(query.params).toMatchObject({});
  });

  it("Generates a single queries when givin attributes data", () => {
    const { queries } = generateCreateEntityRQL(mockBlogSchema, "Blog", {
      title: "test",
      description: "desc",
    });
    expect(queries).toHaveLength(1);
  });

  it("Generates two queries when giving attributes data and a relation", () => {
    const { queries } = generateCreateEntityRQL(mockBlogSchema, "BlogEntry", {
      title: "test",
      content: "content",
      entry_of: 5,
    });
    expect(queries).toHaveLength(2);
  });

  it("Generates the correct query to create without data", () => {
    const { queries } = generateCreateEntityRQL(mockBlogSchema, "Blog");
    const query = queries[0];
    expect(query.rql).toBe("INSERT Blog X");
    expect(query.params).toMatchObject({});
  });

  it("Generates the correct query to create with attributes data", () => {
    const { queries } = generateCreateEntityRQL(mockBlogSchema, "Blog", {
      title: "test",
      description: "desc",
    });
    const createQuery = queries[0];
    expect(createQuery.rql).toBe(
      "INSERT Blog X: X title %(title)s, X description %(description)s"
    );
    expect(createQuery.params).toMatchObject({
      title: "test",
      description: "desc",
    });
  });

  it("Generates the correct queries to create with relation data", () => {
    const { queries } = generateCreateEntityRQL(mockBlogSchema, "BlogEntry", {
      entry_of: 5,
    });
    const createQuery = queries[0];
    expect(createQuery.rql).toBe("INSERT BlogEntry X");
    expect(createQuery.params).toMatchObject({});
    const relationQuery = queries[1];
    expect(relationQuery.rql).toBe(
      "SET X entry_of O WHERE O eid %(objectEid)s, X eid %(subjectEid)s"
    );
    expect(relationQuery.params).toMatchObject({
      subjectEid: { column: 0, row: 0 },
      objectEid: 5,
    });
  });

  it("Generates the correct queries to create with attributes and relation data", () => {
    const { queries } = generateCreateEntityRQL(mockBlogSchema, "BlogEntry", {
      title: "test",
      content: "content",
      entry_of: 5,
    });
    const createQuery = queries[0];
    expect(createQuery.rql).toBe(
      "INSERT BlogEntry X: X title %(title)s, X content %(content)s"
    );
    expect(createQuery.params).toMatchObject({
      title: "test",
      content: "content",
    });
    const relationQuery = queries[1];
    expect(relationQuery.rql).toBe(
      "SET X entry_of O WHERE O eid %(objectEid)s, X eid %(subjectEid)s"
    );
    expect(relationQuery.params).toMatchObject({
      subjectEid: { column: 0, row: 0 },
      objectEid: 5,
    });
  });

  it("Throws when given an unknown relation", () => {
    expect(() => {
      generateCreateEntityRQL(mockBlogSchema, "BlogEntry", {
        title: "test",
        content: "content",
        entry_of: 5,
        error: 5,
      });
    }).toThrowError("error does not exist in entity type BlogEntry");
  });
});
