import { generateCountEntitiesRQL, generateSelectEntitiesRQL } from "../get";
import { RQLQueryRestrictions } from "../utils/RqlQueryRestrictions";
import { mockBlogSchema, mockRelationSchema } from "../utils/test_utils";

describe("generateSelectEntitiesRQL", () => {
  it("Generates a single query for one attribute", () => {
    const queries = generateSelectEntitiesRQL(mockBlogSchema, "BlogEntry", {
      relations: ["title"],
      where: { type: "values", values: { eid: 5 } },
    });
    expect(queries).toHaveLength(1);
  });

  it("Generates a single query for several attributes and relations", () => {
    const queries = generateSelectEntitiesRQL(mockBlogSchema, "BlogEntry", {
      relations: ["title", "content", "entry_of"],
      where: { eid: 5 },
    });
    expect(queries).toHaveLength(1);
  });

  it("Generates the correct query for one attribute", () => {
    const queries = generateSelectEntitiesRQL(mockBlogSchema, "BlogEntry", {
      relations: ["title"],
      where: { eid: 5 },
    });
    expect(queries[0].rql).toBe(
      "Any X, ATTR_TITLE GROUPBY X, ATTR_TITLE WHERE X is BlogEntry, X title ATTR_TITLE, X eid %(eid)s"
    );
    expect(queries[0].params).toMatchObject({
      eid: 5,
    });
  });

  it("Generates the correct query for subject relations with 1? cardinality", () => {
    const queries = generateSelectEntitiesRQL(
      mockRelationSchema,
      "TestSubject",
      {
        relations: ["relation_one_one", "relation_question_one"],
        where: { eid: 5 },
      }
    );
    expect(queries[0].rql).toBe(
      "Any X, REL_RELATION_ONE_ONE, REL_RELATION_QUESTION_ONE GROUPBY X, REL_RELATION_ONE_ONE, REL_RELATION_QUESTION_ONE WHERE X is TestSubject, X relation_one_one REL_RELATION_ONE_ONE, X relation_question_one REL_RELATION_QUESTION_ONE, X eid %(eid)s"
    );
    expect(queries[0].params).toMatchObject({
      eid: 5,
    });
  });

  it("Generates the correct query for subject relations with 1? cardinality", () => {
    const queries = generateSelectEntitiesRQL(
      mockRelationSchema,
      "TestSubject",
      {
        relations: ["relation_one_one", "relation_question_one"],
        where: { eid: 5 },
      }
    );
    expect(queries[0].rql).toBe(
      "Any X, REL_RELATION_ONE_ONE, REL_RELATION_QUESTION_ONE GROUPBY X, REL_RELATION_ONE_ONE, REL_RELATION_QUESTION_ONE WHERE X is TestSubject, X relation_one_one REL_RELATION_ONE_ONE, X relation_question_one REL_RELATION_QUESTION_ONE, X eid %(eid)s"
    );
    expect(queries[0].params).toMatchObject({
      eid: 5,
    });
  });

  it("Generates the correct query for subject relations with +* cardinality", () => {
    const queries = generateSelectEntitiesRQL(
      mockRelationSchema,
      "TestSubject",
      {
        relations: ["relation_plus_one", "relation_star_one"],
        where: { eid: 5 },
      }
    );
    expect(queries[0].rql).toBe(
      "Any X, GROUP_CONCAT(REL_RELATION_PLUS_ONE), GROUP_CONCAT(REL_RELATION_STAR_ONE) GROUPBY X WHERE X is TestSubject, X relation_plus_one REL_RELATION_PLUS_ONE, X relation_star_one REL_RELATION_STAR_ONE, X eid %(eid)s"
    );
    expect(queries[0].params).toMatchObject({
      eid: 5,
    });
  });

  it("Generates the correct query for object relations with 1? cardinality", () => {
    const queries = generateSelectEntitiesRQL(
      mockRelationSchema,
      "TestObject",
      {
        relations: [
          "reverse_relation_one_one",
          "reverse_relation_one_question",
        ],
        where: { eid: 5 },
      }
    );
    expect(queries[0].rql).toBe(
      "Any X, REL_REV_RELATION_ONE_ONE, REL_REV_RELATION_ONE_QUESTION GROUPBY X, REL_REV_RELATION_ONE_ONE, REL_REV_RELATION_ONE_QUESTION WHERE X is TestObject, REL_REV_RELATION_ONE_ONE relation_one_one X, REL_REV_RELATION_ONE_QUESTION relation_one_question X, X eid %(eid)s"
    );
    expect(queries[0].params).toMatchObject({
      eid: 5,
    });
  });

  it("Generates the correct query for object relations with +* cardinality", () => {
    const queries = generateSelectEntitiesRQL(
      mockRelationSchema,
      "TestObject",
      {
        relations: ["reverse_relation_one_plus", "reverse_relation_one_star"],
        where: { eid: 5 },
      }
    );
    expect(queries[0].rql).toBe(
      "Any X, GROUP_CONCAT(REL_REV_RELATION_ONE_PLUS), GROUP_CONCAT(REL_REV_RELATION_ONE_STAR) GROUPBY X WHERE X is TestObject, REL_REV_RELATION_ONE_PLUS relation_one_plus X, REL_REV_RELATION_ONE_STAR relation_one_star X, X eid %(eid)s"
    );
    expect(queries[0].params).toMatchObject({
      eid: 5,
    });
  });

  it("Generates the correct query for several attribute and relations", () => {
    const queries = generateSelectEntitiesRQL(mockBlogSchema, "BlogEntry", {
      relations: ["title", "content", "entry_of"],
      where: { eid: 5 },
    });
    expect(queries[0].rql).toBe(
      "Any X, ATTR_TITLE, ATTR_CONTENT, GROUP_CONCAT(REL_ENTRY_OF) GROUPBY X, ATTR_TITLE, ATTR_CONTENT WHERE X is BlogEntry, X title ATTR_TITLE, X content ATTR_CONTENT, X entry_of REL_ENTRY_OF, X eid %(eid)s"
    );
    expect(queries[0].params).toMatchObject({
      eid: 5,
    });
  });

  it("Generates one query", () => {
    const queries = generateSelectEntitiesRQL(mockBlogSchema, "BlogEntry", {
      limit: 10,
      offset: 0,
      orderBy: [{ relationName: "title", ascending: true }],
      relations: ["title"],
    });
    expect(queries).toHaveLength(1);
  });

  it("Generates the correct query for one attribute without filter", () => {
    const queries = generateSelectEntitiesRQL(mockBlogSchema, "BlogEntry", {
      limit: 10,
      offset: 0,
      orderBy: [{ relationName: "title", ascending: true }],
      relations: ["title"],
    });
    expect(queries[0].rql).toBe(
      "Any X, ATTR_TITLE GROUPBY X, ATTR_TITLE ORDERBY ATTR_TITLE ASC LIMIT 10 OFFSET 0 WHERE X is BlogEntry, X title ATTR_TITLE"
    );
    expect(queries[0].params).toMatchObject({});
  });

  it("Generates the correct query for one attribute with filter", () => {
    const queries = generateSelectEntitiesRQL(mockBlogSchema, "BlogEntry", {
      limit: 10,
      offset: 0,
      orderBy: [{ relationName: "title", ascending: true }],
      where: new RQLQueryRestrictions("X test %(test_number)s", {
        test_number: 5,
      }),
      relations: ["title"],
    });
    expect(queries[0].rql).toBe(
      "Any X, ATTR_TITLE GROUPBY X, ATTR_TITLE ORDERBY ATTR_TITLE ASC LIMIT 10 OFFSET 0 WHERE X is BlogEntry, X title ATTR_TITLE, X test %(test_number)s"
    );
    expect(queries[0].params).toMatchObject({ test_number: 5 });
  });

  it("Generates the correct query for several attributes and relations", () => {
    const queries = generateSelectEntitiesRQL(mockBlogSchema, "BlogEntry", {
      limit: 10,
      offset: 0,
      orderBy: [{ relationName: "title", ascending: true }],
      relations: ["title", "content", "entry_of"],
    });
    expect(queries[0].rql).toBe(
      "Any X, ATTR_TITLE, ATTR_CONTENT, GROUP_CONCAT(REL_ENTRY_OF) GROUPBY X, ATTR_TITLE, ATTR_CONTENT ORDERBY ATTR_TITLE ASC LIMIT 10 OFFSET 0 WHERE X is BlogEntry, X title ATTR_TITLE, X content ATTR_CONTENT, X entry_of REL_ENTRY_OF"
    );
    expect(queries[0].params).toMatchObject({});
  });
});

describe("generateCountEntitiesRQL", () => {
  it("Generates one query", () => {
    const queries = generateCountEntitiesRQL(mockBlogSchema, "BlogEntry", {});
    expect(queries).toHaveLength(1);
  });

  it("Generates the correct query without filter", () => {
    const queries = generateCountEntitiesRQL(mockBlogSchema, "BlogEntry", {});
    expect(queries[0].rql).toBe("Any Count(X) WHERE X is BlogEntry");
    expect(queries[0].params).toMatchObject({});
  });

  it("Generates the correct query with filter", () => {
    const queries = generateCountEntitiesRQL(mockBlogSchema, "BlogEntry", {
      where: new RQLQueryRestrictions("X test %(test_number)s", {
        test_number: 5,
      }),
    });
    expect(queries[0].rql).toBe(
      "Any Count(X) WHERE X is BlogEntry, X test %(test_number)s"
    );
    expect(queries[0].params).toMatchObject({ test_number: 5 });
  });
});
