import { RQLQueryCellRef } from "@cubicweb/client";
import {
  generateDeleteRelationByEidsRQL,
  generateSetAttributesByEidRQL,
  generateSetRelationByEidsRQL,
} from "../update";
import { mockBlogSchema } from "../utils/test_utils";

describe("generateSetAttributesByEidRQL", () => {
  it("Generates a single query for one attribute", () => {
    const queries = generateSetAttributesByEidRQL(mockBlogSchema, "Blog", 5, {
      title: "test",
    });
    expect(queries).toHaveLength(1);
  });

  it("Generates a single query for multiple attributes", () => {
    const queries = generateSetAttributesByEidRQL(mockBlogSchema, "Blog", 5, {
      title: "test",
      description: "desc",
    });
    expect(queries).toHaveLength(1);
  });

  it("Throws an error when no attribute is given", () => {
    expect(() =>
      generateSetAttributesByEidRQL(mockBlogSchema, "Blog", 5, {})
    ).toThrowError("No attributes to set");
  });

  it("Generates a query with the right RQL and params for one attribute", () => {
    const queries = generateSetAttributesByEidRQL(mockBlogSchema, "Blog", 5, {
      title: "test",
    });
    expect(queries[0].rql).toBe(
      "SET X title %(title)s WHERE X is Blog, X eid %(eid)s"
    );
    expect(queries[0].params).toMatchObject({
      eid: 5,
      title: "test",
    });
  });

  it("Generates a query with the right RQL and params for one attribute using an eid ref", () => {
    const ref = new RQLQueryCellRef("test", 0, 0);
    const queries = generateSetAttributesByEidRQL(mockBlogSchema, "Blog", ref, {
      title: "test",
    });
    expect(queries[0].rql).toBe(
      "SET X title %(title)s WHERE X is Blog, X eid %(eid)s"
    );
    expect(queries[0].params).toMatchObject({
      eid: ref,
      title: "test",
    });
  });

  it("Generates a query with the right RQL and params for several attributes", () => {
    const queries = generateSetAttributesByEidRQL(mockBlogSchema, "Blog", 5, {
      title: "test",
      description: "desc",
    });
    expect(queries[0].rql).toBe(
      "SET X title %(title)s, X description %(description)s WHERE X is Blog, X eid %(eid)s"
    );
    expect(queries[0].params).toMatchObject({
      eid: 5,
      title: "test",
      description: "desc",
    });
  });
});

describe("generateSetRelationByEidsRQL", () => {
  it("Throws an error when no subject eid is given", () => {
    expect(() => generateSetRelationByEidsRQL([], "test", [3])).toThrowError(
      "No subject entity given"
    );
  });

  it("Throws an error when no obec object eid is given", () => {
    expect(() => generateSetRelationByEidsRQL([1], "test", [])).toThrowError(
      "No object entity given"
    );
  });

  it("Throws an error when relation type is invalid", () => {
    expect(() => generateSetRelationByEidsRQL([1], "", [3])).toThrowError(
      "Invalid relation name"
    );
  });

  it("Generates a single query for a single subject and object eid", () => {
    const queries = generateSetRelationByEidsRQL([1], "test", [3]);
    expect(queries).toHaveLength(1);
  });

  it("Generates two queries for two subject eid", () => {
    const queries = generateSetRelationByEidsRQL([1, 2], "test", [3]);
    expect(queries).toHaveLength(2);
  });

  it("Generates two queries for two object eid", () => {
    const queries = generateSetRelationByEidsRQL([1], "test", [3, 4]);
    expect(queries).toHaveLength(2);
  });

  it("Generates a four queries for two subject and object eid", () => {
    const queries = generateSetRelationByEidsRQL([1, 2], "test", [3, 4]);
    expect(queries).toHaveLength(4);
  });

  it("Generates the correct query for a single subject and object eid", () => {
    const queries = generateSetRelationByEidsRQL([1], "test", [3]);
    const query = queries[0];
    expect(query.rql).toBe(
      "SET X test O WHERE O eid %(objectEid)s, X eid %(subjectEid)s"
    );
    expect(query.params).toMatchObject({
      objectEid: 3,
      subjectEid: 1,
    });
  });

  it("Generates the correct query for a single subject and object eid ref", () => {
    const ref1 = new RQLQueryCellRef("test1", 0, 0);
    const ref2 = new RQLQueryCellRef("test2", 0, 0);
    const queries = generateSetRelationByEidsRQL([ref1], "test", [ref2]);
    const query = queries[0];
    expect(query.rql).toBe(
      "SET X test O WHERE O eid %(objectEid)s, X eid %(subjectEid)s"
    );
    expect(query.params).toMatchObject({
      objectEid: ref2,
      subjectEid: ref1,
    });
  });
});

describe("pushDeleteRelations", () => {
  it("Throws an error when no subject eid is given", () => {
    expect(() => generateDeleteRelationByEidsRQL([], "test", [3])).toThrowError(
      "No subject entity given"
    );
  });

  it("Throws an error when no obec object eid is given", () => {
    expect(() => generateDeleteRelationByEidsRQL([1], "test", [])).toThrowError(
      "No object entity given"
    );
  });

  it("Throws an error when relation type is invalid", () => {
    expect(() => generateDeleteRelationByEidsRQL([1], "", [3])).toThrowError(
      "Invalid relation name"
    );
  });

  it("Generates a single query for a single subject and object eid", () => {
    const queries = generateDeleteRelationByEidsRQL([1], "test", [3]);
    expect(queries).toHaveLength(1);
  });

  it("Generates two queries for two subject eid", () => {
    const queries = generateDeleteRelationByEidsRQL([1, 2], "test", [3]);
    expect(queries).toHaveLength(2);
  });

  it("Generates two queries for two object eid", () => {
    const queries = generateDeleteRelationByEidsRQL([1], "test", [3, 4]);
    expect(queries).toHaveLength(2);
  });

  it("Generates a four queries for two subject and object eid", () => {
    const queries = generateDeleteRelationByEidsRQL([1, 2], "test", [3, 4]);
    expect(queries).toHaveLength(4);
  });

  it("Generates the correct query for a single subject and object eid", () => {
    const queries = generateDeleteRelationByEidsRQL([1], "test", [3]);
    const query = queries[0];
    expect(query.rql).toBe(
      "DELETE X test O WHERE O eid %(objectEid)s, X eid %(subjectEid)s"
    );
    expect(query.params).toMatchObject({
      objectEid: 3,
      subjectEid: 1,
    });
  });

  it("Generates the correct query for a single subject and object eid ref", () => {
    const ref1 = new RQLQueryCellRef("test1", 0, 0);
    const ref2 = new RQLQueryCellRef("test2", 0, 0);
    const queries = generateDeleteRelationByEidsRQL([ref1], "test", [ref2]);
    const query = queries[0];
    expect(query.rql).toBe(
      "DELETE X test O WHERE O eid %(objectEid)s, X eid %(subjectEid)s"
    );
    expect(query.params).toMatchObject({
      objectEid: ref2,
      subjectEid: ref1,
    });
  });
});
