import {
  attributeToVariable,
  getAttributesUpdateClause,
  getGroupByClause,
  getRestrictionsClause,
  getSelectionClause,
  getRelationVariable,
  objectRelationToVariable,
  getRelationsSchemata,
  subjectRelationToVariable,
} from "../../utils/rql";
import { getLooseSchema } from "@cubicweb/client";
import {
  BLOG_ENTRY_ATTRIBUTES,
  BLOG_ENTRY_SUBJECTS,
  BLOG_OBJECTS,
  TEST_OBJECT_OBJECTS,
  TEST_SUBJECT_SUBJECTS,
  mockBlogSchema,
  mockRelationSchema,
} from "../../utils/test_utils";

describe("attributeToVariable", () => {
  test("Generates with valid name", () => {
    expect(attributeToVariable("attribute_name")).toBe("ATTR_ATTRIBUTE_NAME");
  });
  test("Throws error on empty name", () => {
    expect(() => attributeToVariable("")).toThrow("Empty attribute name");
  });
});

describe("subjectRelationToVariable", () => {
  test("Generates with valid name", () => {
    expect(subjectRelationToVariable("subject_relation_name")).toBe(
      "REL_SUBJECT_RELATION_NAME"
    );
  });
  test("Throws error on empty name", () => {
    expect(() => subjectRelationToVariable("")).toThrow(
      "Empty subject relation name"
    );
  });
});

describe("objectRelationToVariable", () => {
  test("Generates with valid name", () => {
    expect(objectRelationToVariable("object_relation_name")).toBe(
      "REL_REV_OBJECT_RELATION_NAME"
    );
  });
  test("Throws error on empty name", () => {
    expect(() => objectRelationToVariable("")).toThrow(
      "Empty object relation name"
    );
  });
});

describe("getRelationsSchemata", () => {
  const blogEntrySchema =
    getLooseSchema(mockBlogSchema).getEntity("BlogEntry")!;
  const blogSchema = getLooseSchema(mockBlogSchema).getEntity("Blog")!;
  test("Returns all attributes and subject when not specified", () => {
    const resolvedRelations = getRelationsSchemata(blogEntrySchema);
    expect(resolvedRelations.attributes).toHaveLength(3);
    expect(resolvedRelations.subjectRelations).toHaveLength(2);
    expect(resolvedRelations.objectRelations).toHaveLength(0);
    expect(resolvedRelations.attributes).toMatchObject(
      blogEntrySchema.attributes
    );
    expect(resolvedRelations.subjectRelations).toMatchObject(
      blogEntrySchema.relationDefinitions.subject
    );
    expect(resolvedRelations.objectRelations).toMatchObject([]);
  });

  test("Returns only attributes or relations specified", () => {
    const resolvedRelations = getRelationsSchemata(blogEntrySchema, [
      "title",
      "entry_of",
    ]);
    expect(resolvedRelations.attributes).toHaveLength(1);
    expect(resolvedRelations.subjectRelations).toHaveLength(1);
    expect(resolvedRelations.objectRelations).toHaveLength(0);
    expect(resolvedRelations.attributes).toMatchObject([
      blogEntrySchema.attributes.find((r) => r.type === "title"),
    ]);
    expect(resolvedRelations.subjectRelations).toMatchObject([
      blogEntrySchema.relationDefinitions.subject.find(
        (r) => r.type === "entry_of"
      ),
    ]);
    expect(resolvedRelations.objectRelations).toMatchObject([]);
  });

  test("Can handle reverse relations", () => {
    const resolvedRelations = getRelationsSchemata(blogSchema, [
      "reverse_entry_of",
    ]);
    expect(resolvedRelations.attributes).toHaveLength(0);
    expect(resolvedRelations.subjectRelations).toHaveLength(0);
    expect(resolvedRelations.objectRelations).toHaveLength(1);
    expect(resolvedRelations.attributes).toMatchObject([]);
    expect(resolvedRelations.subjectRelations).toMatchObject([]);
    expect(resolvedRelations.objectRelations).toMatchObject([
      blogSchema.relationDefinitions.object.find((r) => r.type === "entry_of"),
    ]);
  });
});

describe("getSelectionClause", () => {
  const testSubjectSchema =
    getLooseSchema(mockRelationSchema).getEntity("TestSubject")!;
  const testObjectSchema =
    getLooseSchema(mockRelationSchema).getEntity("TestObject")!;
  test("Only returns X when no relation is given", () => {
    expect(getSelectionClause([], [], [])).toBe("X");
  });

  test("Returns attributes variables", () => {
    expect(getSelectionClause(BLOG_ENTRY_ATTRIBUTES, [], [])).toBe(
      "X, ATTR_CONTENT, ATTR_CONTENT_FORMAT, ATTR_TITLE"
    );
  });

  test("Returns subject relation variables with cardinality 1?", () => {
    const relation_one_one = testSubjectSchema.matchRelationDefinitions({
      type: "relation_one_one",
    })[0];
    const relation_question_one = testSubjectSchema.matchRelationDefinitions({
      type: "relation_question_one",
    })[0];
    expect(
      getSelectionClause([], [relation_one_one, relation_question_one], [])
    ).toBe("X, REL_RELATION_ONE_ONE, REL_RELATION_QUESTION_ONE");
  });

  test("Returns subject relation variables with cardinality +* with GROUP_CONCAT", () => {
    const relation_plus_one = testSubjectSchema.matchRelationDefinitions({
      type: "relation_plus_one",
    })[0];
    const relation_star_one = testSubjectSchema.matchRelationDefinitions({
      type: "relation_star_one",
    })[0];
    expect(
      getSelectionClause([], [relation_plus_one, relation_star_one], [])
    ).toBe(
      "X, GROUP_CONCAT(REL_RELATION_PLUS_ONE), GROUP_CONCAT(REL_RELATION_STAR_ONE)"
    );
  });

  test("Returns object relation variables with cardinality 1?", () => {
    const relation_one_one = testObjectSchema.matchRelationDefinitions({
      type: "relation_one_one",
    })[0];
    const relation_one_question = testObjectSchema.matchRelationDefinitions({
      type: "relation_one_question",
    })[0];
    expect(
      getSelectionClause([], [], [relation_one_one, relation_one_question])
    ).toBe("X, REL_REV_RELATION_ONE_ONE, REL_REV_RELATION_ONE_QUESTION");
  });

  test("Returns object relation variables with cardinality +* with GROUP_CONCAT", () => {
    const relation_one_plus = testObjectSchema.matchRelationDefinitions({
      type: "relation_one_plus",
    })[0];
    const relation_one_star = testObjectSchema.matchRelationDefinitions({
      type: "relation_one_star",
    })[0];
    expect(
      getSelectionClause([], [], [relation_one_plus, relation_one_star])
    ).toBe(
      "X, GROUP_CONCAT(REL_REV_RELATION_ONE_PLUS), GROUP_CONCAT(REL_REV_RELATION_ONE_STAR)"
    );
  });
});

describe("getRestrictionsClause", () => {
  test("Returns nothing with no relations", () => {
    expect(getRestrictionsClause([], [], [])).toBe("");
  });

  test("Returns only attributes", () => {
    expect(getRestrictionsClause(BLOG_ENTRY_ATTRIBUTES, [], [])).toBe(
      "X content ATTR_CONTENT, X content_format ATTR_CONTENT_FORMAT, X title ATTR_TITLE"
    );
  });

  test("Returns only subjects", () => {
    expect(getRestrictionsClause([], BLOG_ENTRY_SUBJECTS, [])).toBe(
      "X creator REL_CREATOR, X entry_of REL_ENTRY_OF"
    );
  });

  test("Returns only objects", () => {
    expect(getRestrictionsClause([], [], BLOG_OBJECTS)).toBe(
      "REL_REV_ENTRY_OF entry_of X"
    );
  });

  test("Returns all objects", () => {
    expect(
      getRestrictionsClause(
        BLOG_ENTRY_ATTRIBUTES,
        BLOG_ENTRY_SUBJECTS,
        BLOG_OBJECTS
      )
    ).toBe(
      "X content ATTR_CONTENT, X content_format ATTR_CONTENT_FORMAT, X title ATTR_TITLE, X creator REL_CREATOR, X entry_of REL_ENTRY_OF, REL_REV_ENTRY_OF entry_of X"
    );
  });
});

describe("getGroupByClause", () => {
  test("Returns GROUPBY X with no relations", () => {
    expect(getGroupByClause([], [], [])).toBe("GROUPBY X");
  });

  test("Returns attributes", () => {
    expect(getGroupByClause(BLOG_ENTRY_ATTRIBUTES, [], [])).toBe(
      "GROUPBY X, ATTR_CONTENT, ATTR_CONTENT_FORMAT, ATTR_TITLE"
    );
  });

  const subjects = TEST_SUBJECT_SUBJECTS.filter(
    (r) =>
      r.type === "relation_one_one" ||
      r.type === "relation_question_one" ||
      r.type === "relation_plus_one" ||
      r.type === "relation_star_one"
  );

  test("Returns only subjects with cardinality 1?", () => {
    expect(getGroupByClause([], subjects, [])).toBe(
      "GROUPBY X, REL_RELATION_ONE_ONE, REL_RELATION_QUESTION_ONE"
    );
  });

  const objects = TEST_OBJECT_OBJECTS.filter(
    (r) =>
      r.type === "relation_one_one" ||
      r.type === "relation_one_question" ||
      r.type === "relation_one_plus" ||
      r.type === "relation_one_star"
  );

  test("Returns only subjects with cardinality 1?", () => {
    expect(getGroupByClause([], [], objects)).toBe(
      "GROUPBY X, REL_REV_RELATION_ONE_ONE, REL_REV_RELATION_ONE_QUESTION"
    );
  });
});

describe("getRelationVariable", () => {
  const blogEntrySchema =
    getLooseSchema(mockBlogSchema).getEntity("BlogEntry")!;
  const testSubjectSchema =
    getLooseSchema(mockRelationSchema).getEntity("TestSubject")!;
  const testObjectSchema =
    getLooseSchema(mockRelationSchema).getEntity("TestObject")!;
  test("Throws an error when no relation given", () => {
    expect(() => getRelationVariable("test", [], [], [])).toThrowError();
  });

  test("Throws an error when relation type not found", () => {
    expect(() =>
      getRelationVariable("test", blogEntrySchema.attributes, [], [])
    ).toThrowError();
  });

  test("Returns the attribute variable", () => {
    expect(
      getRelationVariable("title", blogEntrySchema.attributes, [], [])
    ).toBe("ATTR_TITLE");
  });

  test("Returns the subject relation variable", () => {
    expect(
      getRelationVariable(
        "relation_one_one",
        [],
        testSubjectSchema.relationDefinitions.subject,
        []
      )
    ).toBe("REL_RELATION_ONE_ONE");
  });

  test("Returns the object relation variable", () => {
    expect(
      getRelationVariable(
        "relation_one_one",
        [],
        [],
        testObjectSchema.relationDefinitions.object
      )
    ).toBe("REL_REV_RELATION_ONE_ONE");
  });
});

describe("getAttributesUpdateClause", () => {
  const blogEntrySchema =
    getLooseSchema(mockBlogSchema).getEntity("BlogEntry")!;

  test("Returns nothing when no attribute data", () => {
    expect(getAttributesUpdateClause(blogEntrySchema, {})).toStrictEqual({
      attributesUpdates: "",
      params: {},
    });
  });

  test("Returns the right attributes and values", () => {
    expect(
      getAttributesUpdateClause(blogEntrySchema, {
        title: "test_title",
        content: "test_content",
      })
    ).toStrictEqual({
      attributesUpdates: "X title %(title)s, X content %(content)s",
      params: {
        title: "test_title",
        content: "test_content",
      },
    });
  });

  test("Throws on unknown attributes", () => {
    expect(() => {
      getAttributesUpdateClause(blogEntrySchema, {
        title: "test_title",
        content: "test_content",
        entry_of: 5,
      });
    }).toThrowError("Attribute entry_of is not in the schema of BlogEntry");
  });
});
