import { EntityId } from "./types.js";
import { getRelationsSchemata } from "./rql.js";
import { getEntitySchema } from "./schema.js";
import { EntityData } from "./types.js";
import {
  EntityRawSchemaArray,
  RelationDefinitionRawSchemaArray,
  RelationDefinitionSchema,
  ResultSet,
  ResultSetRow,
  Schema,
  getReverseRelationType,
  isCardinalityMultiple,
} from "@cubicweb/client";

/**
 * Converts a raw RQL result set to a list of JavaScript object for representing entities.
 *
 * The given attributes and relations must be given in the same order as the values in the row.
 *
 * @param row The RQL result set representing the list of entities
 * @param attributes The list of attributes matching the RQL row
 * @param subjectRelations The list of subject reations matching the RQL row
 * @param objectRelations The list of object relations matching the RQL row
 * @returns An array of objects with attributes/relations as keys and their values as value
 *
 * @category Utilities
 */
export const rowsToEntities = (
  rows: ReadonlyArray<ResultSetRow>,
  attributes: ReadonlyArray<RelationDefinitionSchema>,
  subjectRelations: ReadonlyArray<RelationDefinitionSchema>,
  objectRelations: ReadonlyArray<RelationDefinitionSchema>
): Array<EntityData> => {
  return rows.map((row) =>
    rowToEntity(row, attributes, subjectRelations, objectRelations)
  );
};

/**
 * Converts an eid CSV list to a JavaScript int array
 *
 * @param eidsAsCSV Comma separated eid list
 * @returns
 */
const parseEidList = (eidsAsCSV: string) =>
  eidsAsCSV === ""
    ? []
    : eidsAsCSV
        .split(", ")
        .map((eidStr) => Number.parseInt(eidStr, 10))
        .filter((i) => !isNaN(i));

/**
 * Converts a raw RQL row to a JavaScript object for representing an entity.
 *
 * The given attributes and relations must be given in the same order as the values in the row.
 *
 * @param row The RQL result row representing the entity
 * @param attributes The list of attributes matching the RQL row
 * @param subjectRelations The list of subject reations matching the RQL row
 * @param objectRelations The list of object relations matching the RQL row
 * @returns An object with attributes/relations as keys and their values as value
 *
 * @category Utilities
 */
export function rowToEntity(
  row: ResultSetRow,
  attributes: ReadonlyArray<RelationDefinitionSchema>,
  subjectRelations: ReadonlyArray<RelationDefinitionSchema>,
  objectRelations: ReadonlyArray<RelationDefinitionSchema>
): EntityData {
  if (row.length === 0) {
    return {};
  }

  // eid is always the first item in the result
  if (
    row.length !==
    attributes.length + subjectRelations.length + objectRelations.length + 1
  ) {
    throw new Error(
      "Mismatching arguments. " +
        `Got ${row.length - 1} cells (not counting eid in first position) for ${
          attributes.length
        } attributes, ${subjectRelations.length} subject relations and ${
          objectRelations.length
        } object relations.`
    );
  }
  let parsedId = row.getCell(0) as EntityId;
  if (typeof parsedId === "string") {
    parsedId = parseInt(parsedId);
    if (isNaN(parsedId)) {
      parsedId = row.getCell(0) as EntityId;
    }
  }
  const result: EntityData = { eid: parsedId };
  attributes.forEach((r, i) => {
    result[r.type] = row.getCell(i + 1);
  });
  subjectRelations.forEach((r, i) => {
    const value = row.getCell(i + attributes.length + 1);
    result[r.type] =
      isCardinalityMultiple(r.cardinality.subject) && value
        ? parseEidList(value as string)
        : value;
  });
  objectRelations.forEach((r, i) => {
    const value = row.getCell(
      i + attributes.length + subjectRelations.length + 1
    );
    result[getReverseRelationType(r.type)] =
      isCardinalityMultiple(r.cardinality.object) && value
        ? parseEidList(value as string)
        : value;
  });

  return result;
}

export function resultSetToEntity<
  E extends EntityRawSchemaArray,
  R extends RelationDefinitionRawSchemaArray<E>
>(
  schema: Schema<E, R>,
  entityType: string,
  relations: string[],
  resultSet: ResultSet
) {
  const entitySchema = getEntitySchema(schema, entityType);

  const { attributes, subjectRelations, objectRelations } =
    getRelationsSchemata(entitySchema, relations);

  return rowToEntity(
    resultSet.getRow(0),
    attributes,
    subjectRelations,
    objectRelations
  );
}

export function resultSetToEntities<
  E extends EntityRawSchemaArray,
  R extends RelationDefinitionRawSchemaArray<E>
>(
  schema: Schema<E, R>,
  entityType: string,
  resolve: string[] | undefined,
  resultSet: ResultSet
) {
  const entitySchema = getEntitySchema(schema, entityType);

  const { attributes, subjectRelations, objectRelations } =
    getRelationsSchemata(entitySchema, resolve);

  return rowsToEntities(
    Array.from(resultSet),
    attributes,
    subjectRelations,
    objectRelations
  );
}
