import {
  EntitySchema,
  RQLQueryCellRef,
  RQLQueryParams,
  RelationDefinitionSchema,
  getRawRelationType,
  isCardinalityMultiple,
} from "@cubicweb/client";
import { getEntityRelationNames } from "./schema.js";
import { AttributeValue, EntityData } from "./types.js";
import { RQLQueryRestrictions } from "./RqlQueryRestrictions.js";

/**
 * Converts an attribute type to a RQL variable name.
 * All RQL queries generated in this package use this function to create the variable names.
 * Use this to build your own RQL requests or edit ones generated with this package.
 *
 * @param attribute The attribute type to convert
 * @returns A RQL variable name
 *
 * @category RQL Generation
 */
export function attributeToVariable(attribute: string) {
  if (attribute === "") {
    throw new Error("Empty attribute name");
  }
  return `ATTR_${attribute.toUpperCase()}`;
}

/**
 * Converts a subject relation type to a RQL variable name.
 * All RQL queries generated in this package use this function to create the variable names.
 * Use this to build your own RQL requests or edit ones generated with this package.
 *
 * @param relation The subject relation type to convert
 * @returns A RQL variable name
 *
 * @category RQL Generation
 */
export function subjectRelationToVariable(relation: string) {
  if (relation === "") {
    throw new Error("Empty subject relation name");
  }
  return `REL_${relation.toUpperCase()}`;
}

/**
 * Converts an object relation type to a RQL variable name.
 * All RQL queries generated in this package use this function to create the variable names.
 * Use this to build your own RQL requests or edit ones generated with this package.
 *
 * @param relation The object relation type to convert
 * @returns A RQL variable name
 *
 * @category RQL Generation
 */
export function objectRelationToVariable(relation: string) {
  if (relation === "") {
    throw new Error("Empty object relation name");
  }
  return `REL_REV_${relation.toUpperCase()}`;
}

/**
 * Converts a relation or attribute type to a RQL variable name.
 * All RQL queries generated in this package use this function to create the variable names.
 * Use this to build your own RQL requests or edit ones generated with this package.
 *
 * @remarks Object relations must have the reverse prefix. See [@cubicweb/client:getReverseRelationType](https://cubicweb.pages.logilab.fr/cubicwebjs/client/functions/getReverseRelationType.html)
 *
 * @param relationType The relation type to get the variable name from
 * @param attributes The list of attributes to search the relation type in
 * @param subjectRelations The list of subject relations to search the relation type in
 * @param objectRelations The list of object relations to search the relation type in
 * @returns The name of the relation type depending if it represents an attribute, subject or object relation.
 * @throws If the relation type cannot be found in the given relations.
 *
 * @category RQL Generation
 */
export function getRelationVariable(
  relationType: string,
  attributes: Array<RelationDefinitionSchema>,
  subjectRelations: Array<RelationDefinitionSchema>,
  objectRelations: Array<RelationDefinitionSchema>
) {
  if (relationType === "eid") {
    return "X";
  }
  let rel = attributes.find((r) => r.type === relationType);
  if (rel !== undefined) {
    return attributeToVariable(rel.type);
  }
  rel = subjectRelations.find((r) => r.type === relationType);
  if (rel !== undefined) {
    return subjectRelationToVariable(rel.type);
  }
  rel = objectRelations.find(
    (r) => r.type === getRawRelationType(relationType)
  );
  if (rel !== undefined) {
    return objectRelationToVariable(rel.type);
  }
  throw new Error("Could not find relation " + relationType);
}

/**
 * Resolves the list of attributes, subject and object relations from the given array of relation types.
 * Relation types not existing in the given entity type schema will be ignored.
 *
 * If no relation type array is given, then all attributes and subject relations are returned.
 *
 * @remarks ⚠️ Even relations internal to cubicweb such as `created_by` are returned in this case.
 * Use `RELATIONS_AND_ATTRIBUTES_METAMODEL` constant to filter those out.
 *
 * @param entitySchema The entity type schema
 * @param relationsToResolve The list of relation types to resolve. Object relations must have the reverse prefix. See [@cubicweb/client:getReverseRelationType](https://cubicweb.pages.logilab.fr/cubicwebjs/client/functions/getReverseRelationType.html)
 * @returns An object containing the attributes, subject and object relations lists
 *
 * @category RQL Generation
 */
export function getRelationsSchemata(
  entitySchema: EntitySchema,
  relationsToResolve?: Array<string>
) {
  const relationNames: Array<string> = relationsToResolve ?? [];
  if (relationNames.length === 0) {
    relationNames.push(...getEntityRelationNames(entitySchema, "attributes"));
    relationNames.push(...getEntityRelationNames(entitySchema, "subject"));
  }

  const attributes = relationNames
    .map((r) => {
      return entitySchema.attributes.find((a) => a.type === r);
    })
    .filter((r) => r !== undefined) as Array<RelationDefinitionSchema>;

  const subjectRelations = relationNames
    .map((r) => {
      return entitySchema.relationDefinitions.subject.find((a) => a.type === r);
    })
    .filter((r) => r !== undefined) as Array<RelationDefinitionSchema>;

  const objectRelations = relationNames
    .map((r) => {
      return entitySchema.relationDefinitions.object.find(
        (a) => a.type === getRawRelationType(r)
      );
    })
    .filter((r) => r !== undefined) as Array<RelationDefinitionSchema>;
  return { attributes, subjectRelations, objectRelations };
}

/**
 * Generates the selection clause of a RQL query for the given attributes and relations.
 *
 * @remarks The variable `X` will always be part of the selection clause.
 * If you do not supply any relation, `X` will be returned.
 * Relations with cardinality `+` and `*` will be surrounded with [`GROUP_CONCAT`](https://cubicweb.readthedocs.io/en/stable/book/annexes/rql/language.html?highlight=GROUP_CONCAT#aggregate-functions).
 *
 * @example
 * When given the title and content attributes
 * ```
 *           Selection clause
 *    |---------------------------|
 *    |                           |
 * Any X, ATTR_TITLE, ATTR_CONTENT WHERE X eid 1, X title ATTR_TITLE, X content ATTR_CONTENT
 * ```
 *
 * @param attributes The list of attributes
 * @param subjectRelations The list of subject relations
 * @param objectRelations The list of object relations
 * @returns A string representing a selection clause in a RQL query
 *
 * @category RQL Generation
 */
export function getSelectionClause(
  attributes: Array<RelationDefinitionSchema>,
  subjectRelations: Array<RelationDefinitionSchema>,
  objectRelations: Array<RelationDefinitionSchema>
) {
  const attributesSelectionString =
    attributes.length > 0
      ? `, ${attributes.map((r) => attributeToVariable(r.type)).join(", ")}`
      : "";

  const subjectSelectionString =
    subjectRelations.length > 0
      ? `, ${subjectRelations
          .map((r) => {
            const varName = subjectRelationToVariable(r.type);
            if (isCardinalityMultiple(r.cardinality.subject)) {
              return `GROUP_CONCAT(${varName})`;
            }
            return varName;
          })
          .join(", ")}`
      : "";

  const objectSelectionString =
    objectRelations.length > 0
      ? `, ${objectRelations
          .map((r) => {
            const varName = objectRelationToVariable(r.type);
            if (isCardinalityMultiple(r.cardinality.object)) {
              return `GROUP_CONCAT(${varName})`;
            }
            return varName;
          })
          .join(", ")}`
      : "";
  return `X${attributesSelectionString}${subjectSelectionString}${objectSelectionString}`;
}

/**
 * Generates the filter clause of a RQL query for the given data.
 *
 * @remarks Does not handle object relations
 *
 *  * @example
 * When given the title
 * ```
 *                                                     Filter clause
 *                                                  |-----------------|
 *                                                  |                 |
 * Any X, ATTR_CONTENT WHERE X content ATTR_CONTENT, X title %(title)s
 * ```
 *
 * @param data Object with attributes/relations as keys and their value as values
 * @param entitySchema The schema of the entity to filter
 * @returns A RQL query to filter the entity using the data
 */
export function getFilterRestrictionClause(
  data: EntityData,
  entitySchema: EntitySchema
): RQLQueryRestrictions {
  const rqlParts: string[] = [];
  const params: RQLQueryParams = {};

  ["eid", ...getEntityRelationNames(entitySchema, "attributes")].forEach(
    (relationName) => {
      if (Object.keys(data).includes(relationName)) {
        const value = data[relationName];
        if (Array.isArray(value)) {
          rqlParts.push(`X ${relationName} IN (${value.join(", ")},)`);
        } else {
          rqlParts.push(`X ${relationName} %(${relationName})s`);
          params[relationName] = value;
        }
      }
    }
  );

  getEntityRelationNames(entitySchema, "subject").forEach((relationName) => {
    if (Object.keys(data).includes(relationName)) {
      const value = data[relationName];
      if (Array.isArray(value)) {
        rqlParts.push(`X ${relationName} IN (${value.join(", ")},)`);
      } else {
        rqlParts.push(`X ${relationName} %(${relationName})s`);
        params[relationName] = data[relationName];
      }
    }
  });

  // TODO handle object relations

  return new RQLQueryRestrictions(rqlParts.join(", "), params);
}

/**
 * Generates the restriction clause of a RQL query for the given attributes and relations.
 *
 * @remarks The restriction on `X` never be set.
 * If you do not supply any relation, the returned string will be empty.
 *
 * @example
 * When given the title and content attributes
 * ```
 * Any X, ATTR_TITLE, ATTR_CONTENT
 *                            Restriction clause
 *               |------------------------------------------|
 *               |                                          |
 * WHERE X eid 1, X title ATTR_TITLE, X content ATTR_CONTENT
 * ```
 * @param attributes The list of attributes
 * @param subjectRelations The list of subject relations
 * @param objectRelations The list of object relations
 * @returns A string representing a restriction clause in a RQL query
 *
 * @category RQL Generation
 */
export function getRestrictionsClause(
  attributes: Array<RelationDefinitionSchema>,
  subjectRelations: Array<RelationDefinitionSchema>,
  objectRelations: Array<RelationDefinitionSchema>
) {
  const restrictions: string[] = [];
  restrictions.push(
    attributes.length > 0
      ? `${attributes
          .map((r) => `X ${r.type} ${attributeToVariable(r.type)}`)
          .join(", ")}`
      : ""
  );

  restrictions.push(
    subjectRelations.length > 0
      ? `${subjectRelations
          .map((r) => `X ${r.type} ${subjectRelationToVariable(r.type)}`)
          .join(", ")}`
      : ""
  );

  restrictions.push(
    objectRelations.length > 0
      ? `${objectRelations
          .map((r) => `${objectRelationToVariable(r.type)} ${r.type} X`)
          .join(", ")}`
      : ""
  );
  return restrictions.filter((r) => r !== "").join(", ");
}

/**
 * Generates the GROUP BY clause of a RQL query for the given attributes and relations.
 * It will only consider relations of cardinality `1` and `?`.
 * Relations of cardinality `+` or `*` will be ignored.
 *
 * @remarks The group by on `X` will always be set.
 * If you do not supply any relation, `GROUPBY X` will be returned.
 *
 * @example
 * When given the title and content attributes
 * ```
 *                                           Group by clause
 *                                |-----------------------------------|
 *                                |                                   |
 * Any X, ATTR_TITLE, ATTR_CONTENT GROUPBY X, ATTR_TITLE, ATTR_CONTENT
 * WHERE X is BlogEntry, X title ATTR_TITLE, X content ATTR_CONTENT
 * ```
 * @param attributes The list of attributes
 * @param subjectRelations The list of subject relations
 * @param objectRelations The list of object relations
 * @returns A string representing a GROUP BY clause in a RQL query
 *
 * @category RQL Generation
 */
export function getGroupByClause(
  attributes: Array<RelationDefinitionSchema>,
  subjectRelations: Array<RelationDefinitionSchema>,
  objectRelations: Array<RelationDefinitionSchema>
) {
  const groupByArray: Array<string> = [];
  groupByArray.push(
    attributes.map((r) => attributeToVariable(r.type)).join(", ")
  );
  groupByArray.push(
    subjectRelations
      .filter((r) => !isCardinalityMultiple(r.cardinality.subject))
      .map((r) => subjectRelationToVariable(r.type))
      .join(", ")
  );
  groupByArray.push(
    objectRelations
      .filter((r) => !isCardinalityMultiple(r.cardinality.object))
      .map((r) => objectRelationToVariable(r.type))
      .join(", ")
  );

  const filteredGroupByArray = groupByArray.filter((s) => s !== "");

  return filteredGroupByArray.length > 0
    ? `GROUPBY X, ${filteredGroupByArray.join(", ")}`
    : "GROUPBY X";
}

/**
 * Generates the RQL clause to update attributes for a given entity.
 *
 * @remarks Unknown attributes used as keys in the `attributesData` param will be ignored.
 * If no valid attribute is supplied in the data, the returned clause will be empty.
 *
 * @example
 * When given the title and content attributes
 * ```
 *             Attributes update clause
 *    |----------------------------------------|
 *    |                                        |
 * SET X title %(title)s, X content %(content)s WHERE X is BlogEntry, X eid 1
 * ```
 *
 * @param entitySchema The entity schema
 * @param attributesData An object with each attribute type as key and their values as value
 * @returns An object containing the RQL clause and the parameters associated
 * @throws If an attribute does not exist in the schema
 *
 * @category RQL Generation
 */
export function getAttributesUpdateClause(
  entitySchema: EntitySchema,
  attributesData: Record<string, AttributeValue>
) {
  const attributesUpdates: string[] = [];
  const params: Record<string, AttributeValue | RQLQueryCellRef> = {};

  const dataNames = Object.keys(attributesData);
  const schemaNames = getEntityRelationNames(entitySchema, "attributes");
  dataNames.forEach((relationName) => {
    if (!schemaNames.includes(relationName)) {
      throw new Error(
        `Attribute ${relationName} is not in the schema of ${entitySchema.type}`
      );
    }
    attributesUpdates.push(`X ${relationName} %(${relationName})s`);
    params[relationName] = attributesData[relationName];
  });
  return { attributesUpdates: attributesUpdates.join(", "), params };
}
