import {
  EntityRawSchemaArray,
  RelationDefinitionRawSchemaArray,
  RelationDefinitionSchema,
  EntitySchema,
  EntityRole,
  getReverseRelationType,
  Schema,
  getLooseSchema,
} from "@cubicweb/client";
import {
  RELATIONS_TO_IGNORE,
  RELATION_OBJECTS_TO_IGNORE,
} from "./CubicWebMetamodel.js";

/**
 * Ignores relations which should not be shown in react admin.
 * Those also include relations with an incompatible object type.
 *
 * @param relations The original list of relations
 * @returns Filtered list of relations
 *
 * @category Utilities
 */
const getCompatibleRelations = <
  E extends EntityRawSchemaArray,
  R extends RelationDefinitionRawSchemaArray<E>,
  RDef extends R[number]
>(
  relations: RelationDefinitionSchema<E, R, RDef>[]
) =>
  relations.filter(
    (e) =>
      !RELATIONS_TO_IGNORE.includes(e.type) &&
      !RELATION_OBJECTS_TO_IGNORE.includes(e.object.type)
  );

export function getEntitySchema<
  E extends EntityRawSchemaArray,
  R extends RelationDefinitionRawSchemaArray<E>
>(schema: Schema<E, R>, entityType: string): EntitySchema {
  const entitySchema = getLooseSchema(schema).getEntity(entityType);
  if (entitySchema === null) {
    throw new RangeError(`Entity '${entityType}' does not exist in the schema`);
  }
  return entitySchema;
}

/**
 * Retrieves the list of relation names of the given type.
 *
 * @param entitySchema The entity schema abstraction class
 * @param type The type of relations to fetch
 * @returns A list of names for the given relation type
 *
 * @category Utilities
 */
export const getEntityRelationNames = (
  entitySchema: EntitySchema,
  type: "attributes" | EntityRole
): string[] => {
  let relations: RelationDefinitionSchema[] = [];
  switch (type) {
    case "attributes":
      relations = entitySchema.attributes;
      break;
    case "subject":
    case "object":
      relations = entitySchema.relationDefinitions[type];
      break;
  }
  return getCompatibleRelations(relations).map((a) => {
    if (type === "attributes" || type === "subject") {
      return a.type;
    } else {
      return getReverseRelationType(a.type);
    }
  });
};
