import {
  AttributeConstraint,
  CardinalityPair,
  EntityRawSchema,
  EntityRawSchemaArray,
  EntitySchema,
  getLooseSchema,
  RawSchema,
  RelationDefinitionRawSchema,
  RelationDefinitionSchema,
  Schema,
} from "@cubicweb/client";
import {
  getValidationSchemaFromYams,
  applyStringConstraints,
  applyDateConstraints,
  applyNumberConstraints,
  validateDynamicValues,
  CustomYupErrorsTranslations,
} from "../schemaValidator";
import * as yup from "yup";
import { getRelationKey } from "@cubicweb/data-provider";

const TRANSLATIONS: CustomYupErrorsTranslations = {
  required: "required translation",
  dynamic: () => "dynamic translation",
};

async function safeValidate(
  schema: yup.AnySchema,
  input: number | string | object
) {
  try {
    return { success: await schema.validate(input, { strict: true }) };
  } catch (error) {
    if (!yup.ValidationError.isError(error)) {
      throw error;
    }
    return { error };
  }
}

async function validateRelation(
  schema: Schema,
  data: Record<string, number | number[]>
) {
  const yupSchema = getValidationSchemaFromYams(
    // eslint-disable-next-line @typescript-eslint/no-non-null-assertion
    schema.getEntity("TestEntity")!,
    TRANSLATIONS
  );
  return await safeValidate(yupSchema, data);
}

async function validateString(
  str: string,
  attribute: RelationDefinitionSchema
) {
  const parser = applyStringConstraints(attribute);
  return await safeValidate(parser, str);
}

async function validateInt(int: number, attribute: RelationDefinitionSchema) {
  const parser = applyNumberConstraints(attribute, {
    required: "required translation",
    dynamic: () => "dynamic translation",
  });
  return await safeValidate(parser, int);
}

function getAttribute(
  type: string,
  constraints: Array<AttributeConstraint>
): RelationDefinitionSchema {
  return {
    // @ts-expect-error we do not need to define the whole type
    subject: { type: "TestEntity" },
    // @ts-expect-error we do not need to define the whole type
    object: { type },
    type: "label",
    cardinality: {
      subject: "1",
      object: "1",
    },
    final: true,
    constraints,
  };
}

function getStringAttribute(
  constraints: Array<AttributeConstraint>
): RelationDefinitionSchema {
  return getAttribute("String", constraints);
}

function getIntAttribute(
  constraints: Array<AttributeConstraint>
): RelationDefinitionSchema {
  return getAttribute("Int", constraints);
}

function getRelationSchema(
  cardinality: CardinalityPair,
  useMultipleObjects = false
) {
  function _getRelation(
    object: string
  ): RelationDefinitionRawSchema<EntityRawSchemaArray> {
    return {
      subject: "TestEntity",
      object,
      type: "relation",
      cardinality,
      final: false,
      constraints: [],
      description: "",
      options: {},
    };
  }

  const fakeSchema: RawSchema = {
    entities: [
      {
        type: "TestEntity",
        final: false,
        description: "",
      },
      {
        type: "Object1",
        final: false,
        description: "",
      },
      {
        type: "Object2",
        final: false,
        description: "",
      },
      {
        type: "Object3",
        final: false,
        description: "",
      },
    ],
    relations_definitions: useMultipleObjects
      ? [
          _getRelation("Object1"),
          _getRelation("Object2"),
          _getRelation("Object3"),
        ]
      : [_getRelation("Object1")],
  };
  return new Schema(fakeSchema);
}

const personEntity: EntityRawSchema = {
  type: "Person",
  description: "A Person",
  final: false,
};

const stringEntity: EntityRawSchema = {
  type: "String",
  description: "",
  final: true,
};

const intEntity: EntityRawSchema = {
  type: "Int",
  description: "",
  final: true,
};

const nameAttribute: RelationDefinitionRawSchema<EntityRawSchemaArray> = {
  type: "name",
  description: "",
  final: true,
  cardinality: "11",
  subject: "Person",
  object: "String",
  constraints: [
    {
      type: "SizeConstraint",
      value: {
        max: 10,
        min: null,
        msg: null,
      },
    },
  ],
  default: null,
  options: {},
};

const surnameAttribute: RelationDefinitionRawSchema<EntityRawSchemaArray> = {
  type: "surname",
  description: "",
  final: true,
  cardinality: "??",
  subject: "Person",
  object: "String",
  constraints: [],
  default: null,
  options: {},
};

const eidAttribute: RelationDefinitionRawSchema<EntityRawSchemaArray> = {
  type: "eid",
  description: "",
  final: true,
  cardinality: "11",
  subject: "Person",
  object: "Int",
  constraints: [],
  default: null,
  options: {},
};

const nameEidSchema = new Schema({
  entities: [personEntity, stringEntity, intEntity],
  relations_definitions: [eidAttribute, nameAttribute, surnameAttribute],
});

describe("getValidationSchemaFromYams", () => {
  // eslint-disable-next-line @typescript-eslint/no-non-null-assertion
  const entityTypeSchema = nameEidSchema.getEntity("Person")!;

  const OMITTED_FIELDS = ["name"];

  function getExpectedElements(omitFields: boolean) {
    return (
      [
        ...entityTypeSchema.attributes,
        ...entityTypeSchema.relationDefinitions.subject,
      ]
        .filter((r) => !(omitFields && OMITTED_FIELDS.includes(r.type)))
        .map((r) => {
          if (r.final) {
            return r.type;
          }
          return getRelationKey(r, "subject");
        })
        // Sort elements so we can easily compare arrays
        .sort()
    );
  }

  it("Returns a yup object containing all fields", () => {
    const expectedAttributesAndRelations = getExpectedElements(false);
    const validationSchema = getValidationSchemaFromYams(
      entityTypeSchema,
      TRANSLATIONS
    );
    const fields = Object.keys(validationSchema.fields).sort();
    expect(fields).toEqual(expectedAttributesAndRelations);
  });
  it("Returns a yup object without omitted fields", () => {
    const expectedAttributesAndRelations = getExpectedElements(true);
    const validationSchema = getValidationSchemaFromYams(
      entityTypeSchema,
      TRANSLATIONS,
      OMITTED_FIELDS
    );
    const fields = Object.keys(validationSchema.fields).sort();
    expect(fields).toEqual(expectedAttributesAndRelations);
  });
});

describe("schema-validator", () => {
  // eslint-disable-next-line @typescript-eslint/no-non-null-assertion
  const entityTypeSchema = getLooseSchema(nameEidSchema).getEntity("Person")!;

  const OMITTED_FIELDS = ["name"];

  const TestEntity1 = getValidationSchemaFromYams(
    entityTypeSchema,
    TRANSLATIONS
  );

  const TestEntity2 = getValidationSchemaFromYams(
    entityTypeSchema,
    TRANSLATIONS,
    OMITTED_FIELDS
  );

  const rawInput: { eid: number; name: string; surname?: string } = {
    eid: 1,
    name: "test name",
    surname: "test surname",
  };

  it("validates without surname when omitted", async () => {
    const input = { eid: 1 };

    const validatedInput = await safeValidate(TestEntity2, input);
    expect(validatedInput).toHaveProperty("success");
  });

  it("validates with name", async () => {
    const validatedInput = await safeValidate(TestEntity1, rawInput);
    expect(validatedInput).toHaveProperty("success");
  });

  it("fails since name is missing", async () => {
    const input = { eid: 1 };

    const validatedInput = await safeValidate(TestEntity1, input);
    expect(validatedInput.error?.path).toEqual("name");
    expect(validatedInput.error?.type).toEqual("required");
  });

  it("fails since a string is expected", async () => {
    const input = {
      ...rawInput,
      surname: 123,
    };

    const validatedInput = await safeValidate(TestEntity1, input);
    expect(validatedInput.error?.path).toEqual("surname");
    expect(validatedInput.error?.type).toEqual("typeError");
  });

  it("fails with too long string", async () => {
    const input = {
      eid: 1,
      name: "a".repeat(15),
    };
    const validatedInput = await safeValidate(TestEntity1, input);
    expect(validatedInput.error?.path).toEqual("name");
    expect(validatedInput.error?.type).toEqual("max");
  });
});

describe("Relation constraints", () => {
  const _schema = getRelationSchema("11", true);
  const rel1 = _schema.matchRelationDefinitions({
    subject: "TestEntity",
    type: "relation",
    object: "Object1",
  })[0];
  const rel3 = _schema.matchRelationDefinitions({
    subject: "TestEntity",
    type: "relation",
    object: "Object3",
  })[0];
  const object1Key = getRelationKey(rel1, "subject");
  const object3Key = getRelationKey(rel3, "subject");

  describe("Relations with only one object type", () => {
    it("fails when required relation is missing", async () => {
      const schema = getRelationSchema("11", false);
      const validatedInput = await validateRelation(schema, {});
      expect(validatedInput.error?.type).toEqual("required");
      expect(validatedInput.error?.path).toEqual(object1Key);
    });

    it("fails when required relation array is missing", async () => {
      const schema = getRelationSchema("11", false);
      const validatedInput = await validateRelation(schema, {});
      expect(validatedInput.error?.type).toEqual("required");
      expect(validatedInput.error?.path).toEqual(object1Key);
    });

    it("fails when given array instead of number", async () => {
      const schema = getRelationSchema("11", false);
      const validatedInput = await validateRelation(schema, {
        [object1Key]: [1],
      });
      expect(validatedInput.error?.type).toEqual("typeError");
      expect(validatedInput.error?.path).toEqual(object1Key);
    });

    it("fails when given a number instead of an array", async () => {
      const schema = getRelationSchema("+1", false);
      const validatedInput = await validateRelation(schema, {
        [object1Key]: 1,
      });
      expect(validatedInput.error?.type).toEqual("typeError");
      expect(validatedInput.error?.path).toEqual(object1Key);
    });

    it("succeed when required relation is present", async () => {
      const schema = getRelationSchema("11", false);
      const validatedInput = await validateRelation(schema, {
        [object1Key]: 1,
      });
      expect(validatedInput).toHaveProperty("success");
    });

    it("succeed when required relation array is present", async () => {
      const schema = getRelationSchema("+1", false);
      const validatedInput = await validateRelation(schema, {
        [object1Key]: [1],
      });
      expect(validatedInput).toHaveProperty("success");
    });
  });
  describe("Relations with several object types", () => {
    it("fails when all required relations are missing", async () => {
      const schema = getRelationSchema("11", true);
      const validatedInput = await validateRelation(schema, {});
      expect(validatedInput.error?.type).toEqual("required");
      // Will fail on the last object type
      expect(validatedInput.error?.path).toEqual(object3Key);
    });

    it("fails when all required relation arrays are missing", async () => {
      const schema = getRelationSchema("+1", true);
      const validatedInput = await validateRelation(schema, {});
      expect(validatedInput.error?.type).toEqual("required");
      // Will fail on the last object type
      expect(validatedInput.error?.path).toEqual(object3Key);
    });

    it("fails when given array instead of number", async () => {
      const schema = getRelationSchema("11", true);
      const validatedInput = await validateRelation(schema, {
        [object1Key]: [1],
      });
      expect(validatedInput.error?.type).toEqual("typeError");
      expect(validatedInput.error?.path).toEqual(object1Key);
    });

    it("fails when given a number instead of an array", async () => {
      const schema = getRelationSchema("+1", true);
      const validatedInput = await validateRelation(schema, {
        [object1Key]: 1,
      });
      expect(validatedInput.error?.type).toEqual("typeError");
      expect(validatedInput.error?.path).toEqual(object1Key);
    });

    it("succeed when at least one required relation is present", async () => {
      const schema = getRelationSchema("11", true);
      const validatedInput = await validateRelation(schema, {
        [object1Key]: 1,
      });
      expect(validatedInput).toHaveProperty("success");
    });

    it("succeed when at least one required relation array is present", async () => {
      const schema = getRelationSchema("+1", true);
      const validatedInput = await validateRelation(schema, {
        [object1Key]: [1],
      });
      expect(validatedInput).toHaveProperty("success");
    });
  });
});

describe("String constraints", () => {
  const vocabConstraint: Array<AttributeConstraint> = [
    {
      type: "StaticVocabularyConstraint",
      value: {
        msg: null,
        values: [
          "external_temperatures_dry",
          "external_temperatures_wet",
          "external_temperatures_water",
          "external_radiation",
          "external_humidity",
        ],
      },
    },
  ];

  it("fails when vocabulary constraint is not respected", async () => {
    const attribute = getStringAttribute(vocabConstraint);
    const validatedInput = await validateString("coucou", attribute);
    expect(validatedInput).toHaveProperty("error");
  });

  it("success when vocabulary constraint is respected", async () => {
    const attribute = getStringAttribute(vocabConstraint);
    const validatedInput = await validateString(
      "external_temperatures_dry",
      attribute
    );
    expect(validatedInput).toHaveProperty("success");
  });

  const formatConstraint: Array<AttributeConstraint> = [
    {
      type: "FormatConstraint",
      value: {
        msg: null,
        values: ["text/rest", "text/markdown", "text/html", "text/plain"],
      },
    },
  ];

  it("success when format constraint is respected", async () => {
    const attribute = getStringAttribute(formatConstraint);
    const validatedInput = await validateString("text/html", attribute);
    expect(validatedInput).toHaveProperty("success");
  });

  it("fails when format constraint is not respected", async () => {
    const attribute = getStringAttribute(formatConstraint);
    const validatedInput = await validateString("test", attribute);
    expect(validatedInput).toHaveProperty("error");
  });

  it("fails with too long string", async () => {
    const attribute = getStringAttribute([
      { type: "SizeConstraint", value: { min: 0, max: 9, msg: "" } },
    ]);
    const validatedInput = await validateString("a".repeat(10), attribute);
    expect(validatedInput).toHaveProperty("error");
  });

  it("fails with too short string", async () => {
    const attribute = getStringAttribute([
      { type: "SizeConstraint", value: { min: 15, max: null, msg: "" } },
    ]);
    const validatedInput = await validateString("a".repeat(10), attribute);
    expect(validatedInput).toHaveProperty("error");
  });

  const regexpConstraint: Array<AttributeConstraint> = [
    {
      type: "RegexpConstraint",
      value: {
        flags: 0,
        msg: "Frequency must end with s, h or min and start with a number",
        regexp: "^\\d+(?:s|h|min)$",
      },
    },
  ];

  it("fails when regexp does not match", async () => {
    const attribute = getStringAttribute(regexpConstraint);
    const validatedInput = await validateString("a", attribute);
    expect(validatedInput).toHaveProperty("error");
  });

  it("fails when regexp does not match but size is ok", async () => {
    const attribute = getStringAttribute(regexpConstraint);
    const validatedInput = await validateString("3mi", attribute);
    expect(validatedInput).toHaveProperty("error");
  });

  it("succeed when regexp matches", async () => {
    const attribute = getStringAttribute(regexpConstraint);
    const validatedInput = await validateString("6min", attribute);
    expect(validatedInput).toHaveProperty("success");
  });
});

describe("Date constraints", () => {
  it("should succeed when input is greater than static boundary", async () => {
    const attribute = getAttribute("TZDatetime", [
      {
        type: "BoundaryConstraint",
        value: { boundary: 20, op: ">", msg: "" },
      },
    ]);
    const validator = applyDateConstraints(attribute, TRANSLATIONS);
    const validatedInput = await safeValidate(validator, new Date("12/12/12"));
    expect(validatedInput).toHaveProperty("success");
  });
});

describe("Number constraints", () => {
  it("fails when Int is too big", async () => {
    const attribute = getIntAttribute([
      {
        type: "IntervalBoundConstraint",
        value: { minvalue: 0, maxvalue: 9, msg: "" },
      },
    ]);
    const validatedInput = await validateInt(10, attribute);
    expect(validatedInput).toHaveProperty("error");
  });

  it("fails when Int is too small", async () => {
    const attribute = getIntAttribute([
      {
        type: "IntervalBoundConstraint",
        value: { minvalue: 15, maxvalue: null, msg: "" },
      },
    ]);
    const validatedInput = await validateInt(10, attribute);
    expect(validatedInput).toHaveProperty("error");
  });

  describe("BoundaryConstraint", () => {
    it("should succeed when input is greater than static boundary", async () => {
      const attribute = getIntAttribute([
        {
          type: "BoundaryConstraint",
          value: { boundary: 20, op: ">", msg: "" },
        },
      ]);
      const validatedInput = await validateInt(30, attribute);
      expect(validatedInput).toHaveProperty("success");
    });

    it("should error when input is smaller than the static boundary", async () => {
      const attribute = getIntAttribute([
        {
          type: "BoundaryConstraint",
          value: { boundary: 20, op: ">", msg: "" },
        },
      ]);
      const validatedInput = await validateInt(10, attribute);
      expect(validatedInput).toHaveProperty("error");
    });

    const linkedNumbersYams = {
      attributes: [
        {
          type: "count1",
          subject: { type: "TestEntity" },
          object: { type: "Int" },
          cardinality: {
            subject: "1",
            object: "1",
          },
          final: true,
          constraints: [],
        },
        {
          type: "count2",
          subject: { type: "TestEntity" },
          object: { type: "Int" },
          cardinality: {
            subject: "1",
            object: "1",
          },
          final: true,
          constraints: [
            {
              type: "BoundaryConstraint",
              value: {
                boundary: { __attribute__: "count1" },
                op: ">",
              },
            },
          ],
        },
      ],
      relationDefinitions: { subject: [], object: [] },
      // Fake the entity schema for test's sake
    } as unknown as EntitySchema;

    it("should success when input is greater than the other attribute boundary", async () => {
      const rawInput = {
        count1: 30,
        count2: 40,
      };

      const validator = getValidationSchemaFromYams(
        linkedNumbersYams,
        TRANSLATIONS
      );
      const validatedInput = await safeValidate(validator, rawInput);
      expect(validatedInput).toHaveProperty("success");
    });

    it("should error when input is not greater than the other attribute boundary", async () => {
      const rawInput = {
        count1: 30,
        count2: 20,
      };

      const validator = getValidationSchemaFromYams(
        linkedNumbersYams,
        TRANSLATIONS
      );
      const validatedInput = await safeValidate(validator, rawInput);
      expect(validatedInput).toHaveProperty("error");
    });
  });

  describe("validateDynamicValues function", () => {
    const schema = yup.object().shape({
      field1: yup.number().required(),
      field2: yup.number().required(),
    });

    it("should return true when using the greaterThan operator and value is greater", async () => {
      const greaterThanOperator = ">";
      const validatedSchema = schema.shape({
        field1: validateDynamicValues(
          yup.number(),
          "field2",
          greaterThanOperator,
          TRANSLATIONS
        ),
      });
      const validator = await safeValidate(validatedSchema, {
        field1: 5,
        field2: 4,
      });
      expect(validator).toHaveProperty("success");
    });

    it("should errored when using the greaterThan operator and value is not greater", async () => {
      const greaterThanOperator = ">";
      const validatedSchema = schema.shape({
        field1: validateDynamicValues(
          yup.number(),
          "field2",
          greaterThanOperator,
          TRANSLATIONS
        ),
      });
      const validator = await safeValidate(validatedSchema, {
        field1: 4,
        field2: 5,
      });

      expect(validator).toHaveProperty("error");
    });
  });
});
