import {
  AttributeConstraint,
  ConstraintOperator,
  EntityRole,
  EntitySchema,
  getRelationType,
  isCardinalityMultiple,
  isCardinalityRequired,
  RelationDefinitionSchema,
} from "@cubicweb/client";

import * as yup from "yup";
import { getRelationKey } from "@cubicweb/data-provider";

/**
 * Custom errors not handled by yup.setLocale
 *
 * @category Yup
 */
export interface CustomYupErrorsTranslations {
  /** When a relation required is not filled. */
  required: string;
  /**
   * When an attribute does not respect a constraint to another attribute.
   *
   * @param comparisonOperator The comparison operator between the attribute in error and the one linked
   * @param fieldName The linked attribute which the one in error compares to
   * @returns A translated error string
   */
  dynamic: (comparisonOperator: string, fieldName: string) => string;
}

/**
 *
 * If you use {@link useCubicWebForm}, then there is no need to manually use this function.
 *
 * But if you want to write your own hooks, you might still need the yup validation.
 * In that case, you can use this function to generate a Yup object,
 * which you can then use in the the rest of your app.
 *
 * @example
 * Here is an example usage with `react-hook-form`:
 *
 * ```ts
 * import {useForm} from "react-hook-form";
 * import {yupResolver} from "@hookform/resolvers/yup";
 * import {getValidationSchemaFromYams} from "@cubicweb/yup";
 *
 * const entityTypeParser = getValidationSchemaFromYams(
 *     entityTypeSchema,
 *     errorsTranslations,
 *     omittedFields
 * );
 *
 * const {...} = useForm({
 *     resolver: yupResolver(entityTypeParser),
 * });
 * ```
 *
 * @param yams The entity schema containing all the attributes of the entity.
 * @param customErrorsTranslations An object for custom translations
 * @param omittedFields The list of attributes/relations to ignore.
 * @returns The [yup object schema](https://github.com/jquense/yup#api) containing various primitive types to validate.
 *
 * @category Yup
 */
export const getValidationSchemaFromYams = (
  yams: EntitySchema,
  customErrorsTranslations: CustomYupErrorsTranslations,
  omittedFields?: string[]
) => {
  const definePropertyName = (
    relation: RelationDefinitionSchema,
    role: EntityRole
  ) => {
    if (relation.final) {
      return relation.type;
    }
    return getRelationKey(relation, role);
  };

  const attributesAndSubjectRelations = [
    ...yams.attributes,
    ...yams.relationDefinitions.subject,
  ].filter(
    (r) => omittedFields === undefined || !omittedFields.includes(r.type)
  );

  const objectRelations = yams.relationDefinitions.object.filter(
    (r) =>
      omittedFields === undefined ||
      !omittedFields.includes(getRelationType(r.type, "object"))
  );

  const attributesAndSubjectRelationsValidators = Object.fromEntries(
    attributesAndSubjectRelations.map((relation) => [
      definePropertyName(relation, "subject"),
      relationSchemaToYup(relation, "subject", customErrorsTranslations),
    ])
  );

  const objectRelationsValidators = Object.fromEntries(
    objectRelations.map((relation) => [
      definePropertyName(relation, "object"),
      relationSchemaToYup(relation, "object", customErrorsTranslations),
    ])
  );

  return yup.object({
    ...attributesAndSubjectRelationsValidators,
    ...objectRelationsValidators,
  });
};

/**
 * Converts attribute schema to corresponding Yup validation schema
 *
 * @param relDef - The relation definition object
 * @return Yup validation schema
 * @throws Unexpected type of the attribute schema
 */
const relationSchemaToYup = (
  relDef: RelationDefinitionSchema,
  role: EntityRole,
  customErrorsTranslations: CustomYupErrorsTranslations
): yup.BaseSchema => {
  if (relDef.final) {
    switch (relDef.object.type) {
      case "Int":
        return applyNumberConstraints(relDef, customErrorsTranslations);
      case "Float":
        return applyNumberConstraints(relDef, customErrorsTranslations);
      case "String":
        return applyStringConstraints(relDef);
      case "Boolean":
        return yup.boolean();
      case "TZDatetime":
      case "TZDate":
      case "Datetime":
      case "Date":
        return applyDateConstraints(relDef, customErrorsTranslations);
      default:
        console.warn(
          `Unexpected type '${relDef.object.type}' for attribute '${relDef.type}'`
        );
        return yup.mixed();
    }
  }
  return applyRelationRequiredConstraint(
    relDef,
    role,
    customErrorsTranslations
  );
};

export function applyRelationRequiredConstraint(
  relDef: RelationDefinitionSchema,
  role: EntityRole,
  customErrorsTranslations: CustomYupErrorsTranslations
) {
  const baseYupSchema = !isCardinalityMultiple(relDef.cardinality[role])
    ? yup.number()
    : yup.array().of(yup.number());
  if (relDef.isRequired(role)) {
    const sameTypeRelations = relDef[role].relationDefinitions.subject.filter(
      (r) => r.type === relDef.type
    );
    if (sameTypeRelations.length <= 1) {
      return baseYupSchema.required();
    }
    // If there are different object types possible for a single relation,
    // check at least one is set
    const testRelationSet = (testContext: yup.TestContext) => {
      const { parent } = testContext;
      return sameTypeRelations.some((r) => {
        const key = getRelationKey(r, role);
        return parent && parent[key] !== undefined;
      });
    };
    // test function signature is different between number and array
    // so we cannot use baseYupSchema to make a single call
    if (relDef.cardinality[role] === "1") {
      return yup
        .number()
        .test(
          "required",
          customErrorsTranslations.required,
          (_value: number | undefined, testContext: yup.TestContext) =>
            testRelationSet(testContext)
        );
    }
    return yup
      .array()
      .of(yup.number())
      .test(
        "required",
        customErrorsTranslations.required,
        (
          _value: (number | undefined)[] | undefined,
          testContext: yup.TestContext
        ) => testRelationSet(testContext)
      );
  }
  return baseYupSchema.nullable();
}

/**
 * NUMBER CONSTRAINTS
 */

/**

This function applies number constraints to a relation definition.

@param relDef - The relation definition schema to apply the constraints.

@returns Returns a yup NumberSchema with applied constraints.
*/
export function applyNumberConstraints(
  relDef: RelationDefinitionSchema,
  customErrorsTranslations: CustomYupErrorsTranslations
) {
  const numberConstraints = (
    relDef.constraints as AttributeConstraint[]
  ).reduce((acc, current) => {
    if (current.type === "IntervalBoundConstraint") {
      const min = current.value.minvalue;
      if (min !== null && typeof min === "number") {
        acc = acc.min(min);
      }
      const max = current.value.maxvalue;
      if (max !== null && typeof max === "number") {
        acc = acc.max(max);
      }
    }
    if (current.type === "BoundaryConstraint") {
      if (
        current.value.boundary !== null &&
        typeof current.value.boundary === "number"
      ) {
        const { op, boundary } = current.value;
        switch (op) {
          case ">":
            acc = acc.moreThan(boundary);
            break;
          case "<":
            acc = acc.lessThan(boundary);
            break;
          case ">=":
            acc = acc.min(boundary);
            break;
          case "<=":
            acc = acc.max(boundary);
            break;
        }
      }
      if (
        current.value.boundary !== null &&
        // eslint-disable-next-line @typescript-eslint/ban-ts-comment
        // @ts-ignore

        current.value.boundary.__attribute__ !== undefined
      ) {
        acc = validateDynamicValues(
          acc,
          // eslint-disable-next-line @typescript-eslint/ban-ts-comment
          // @ts-ignore
          current.value.boundary.__attribute__,
          current.value.op,
          customErrorsTranslations
        );
      }
    }
    return acc;
  }, yup.number());
  return isCardinalityRequired(relDef.cardinality.subject)
    ? numberConstraints.required()
    : numberConstraints;
}

/**
 * Returns a Yup test function which validates dynamic constraint values for numbers.
 *
 * @param initialSchema - The initial number schema.
 * @param fieldName - The name of the field to validate against.
 * @param comparisonOperator - The comparison operator used to compare the field value with the dynamic value.
 *
 * @returns A Yup test function that is used to validate dynamic constraint values for numbers.
 */
export const validateDynamicValues = (
  initialSchema: yup.NumberSchema,
  fieldName: string,
  comparisonOperator: ConstraintOperator,
  customErrorsTranslations: CustomYupErrorsTranslations
) =>
  initialSchema.test(fieldName, function (value: number | undefined) {
    const { path, parent, createError } = this;
    if (
      value !== undefined &&
      !comparisonOperatorsForNumbers[comparisonOperator](
        value,
        parent[fieldName]
      )
    ) {
      return createError({
        path,
        message: customErrorsTranslations.dynamic(
          comparisonOperator,
          fieldName
        ),
      });
    }
    return true;
  });

const comparisonOperatorsForNumbers: Record<
  ConstraintOperator,
  (a: number, b: number) => boolean
> = {
  "<": (a, b) => a < b,
  ">": (a, b) => a > b,
  ">=": (a, b) => a >= b,
  "<=": (a, b) => a <= b,
};

/**
 * STRING CONSTRAINTS
 */

/**
Adds string constraints to a Yup schema based on the given RelationDefinitionSchema.
@param relDef - The definition of the relation object (with type 'String') from which to derive the validator function.
@return The resulting Yup string validator.
*/
export function applyStringConstraints(relDef: RelationDefinitionSchema) {
  const stringConstraints = (
    relDef.constraints as AttributeConstraint[]
  ).reduce((acc, current) => {
    if (current.type === "RegexpConstraint") {
      if (current.value.regexp !== null) {
        let regex = null;
        try {
          regex = new RegExp(current.value.regexp);
        } catch (e) {
          console.error(
            `Could not parse expression from RegexpConstraint ${regex}`
          );
        }
        if (regex) {
          acc = acc.matches(regex, current.value.msg as string);
        }
      }
    }
    if (current.type === "SizeConstraint") {
      if (current.value.min !== null) {
        acc = acc.min(current.value.min);
      }
      if (current.value.max !== null) {
        acc = acc.max(current.value.max);
      }
    }
    if (
      current.type === "StaticVocabularyConstraint" ||
      current.type === "FormatConstraint"
    ) {
      const vocabulary: (string | undefined)[] = current.value.values;
      if (vocabulary.length !== 0) {
        acc = acc.oneOf(vocabulary);
      }
    }
    return acc;
  }, yup.string());
  return isCardinalityRequired(relDef.cardinality.subject)
    ? stringConstraints.required()
    : stringConstraints;
}

/**
 * DATES CONSTRAINTS
 */

/**
 * Apply constraints to a Yup schema instance for dates based on the given relation definition.
 *
 * @param relDef - The schema of the relation between entities.
 *
 * @returns A Yup schema instance for dates with constraints applied.
 */
export function applyDateConstraints(
  relDef: RelationDefinitionSchema,
  customErrorsTranslations: CustomYupErrorsTranslations
) {
  const dateConstraints = (relDef.constraints as AttributeConstraint[]).reduce(
    (acc, current) => {
      if (current.type === "BoundaryConstraint") {
        if (
          current.value.boundary !== null &&
          typeof current.value.boundary === "string"
        ) {
          const operator = current.value
            .op as SupportedComparisonOperatorsForDates;
          if (operator === "<") {
            acc = acc.max(current.value.boundary);
          } else {
            acc = acc.min(current.value.boundary);
          }
        }
        if (
          current.value.boundary !== null &&
          // eslint-disable-next-line @typescript-eslint/ban-ts-comment
          // @ts-ignore

          current.value.boundary.__attribute__ !== undefined
        ) {
          acc = validateDynamicValuesForDates(
            acc,
            // eslint-disable-next-line @typescript-eslint/ban-ts-comment
            // @ts-ignore

            current.value.boundary.__attribute__,
            current.value.op as SupportedComparisonOperatorsForDates,
            customErrorsTranslations
          );
        }
      }
      return acc;
    },
    yup.date()
  );
  return isCardinalityRequired(relDef.cardinality.subject)
    ? dateConstraints.required()
    : dateConstraints;
}

type SupportedComparisonOperatorsForDates = "<" | ">";

/**
 * A function that compares two dates using the given supported comparison operator for dates
 * @param comparisonOperator - The comparison operator to use (<, >, <= or >=)
 * @param a - The first date to compare
 * @param b - The second date to compare
 * @returns Returns the result of the evaluation and default to true when one the two dates are undefined
 */
const comparisonOperatorsForDates = (
  comparisonOperator: SupportedComparisonOperatorsForDates,
  a: Date | undefined,
  b: Date | undefined
): boolean => {
  if (comparisonOperator === "<" && a !== undefined && b !== undefined) {
    return a.getTime() < b.getTime();
  }
  if (comparisonOperator === ">" && a !== undefined && b !== undefined) {
    return a.getTime() > b.getTime();
  }
  return true;
};

/**
 * Validates dynamic values for dates
 *
 * @param initialSchema - The initial schema to apply the test
 * @param fieldName - The name of the field
 * @param comparisonOperator - Supported date comparison operator
 * @return true if the field is validated, otherwise returns the errors
 */
export const validateDynamicValuesForDates = (
  initialSchema: yup.DateSchema,
  fieldName: string,
  comparisonOperator: SupportedComparisonOperatorsForDates,
  customErrorsTranslations: CustomYupErrorsTranslations
) =>
  initialSchema.test(fieldName, function (value: Date | undefined) {
    const { path, parent, createError } = this;
    if (
      !comparisonOperatorsForDates(comparisonOperator, value, parent[fieldName])
    ) {
      return createError({
        path,
        message: customErrorsTranslations.dynamic(
          comparisonOperator,
          fieldName
        ),
      });
    }
    return true;
  });
